##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
import sys
import h5pyd
import numpy as np

cube_side = 256
USER_NAME = "test_user1"
USER_PASSWD = "test"
ENDPOINT = "http://192.168.99.100:5101"

if len(sys.argv) > 1:
    cube_side = int(sys.argv[1])

filename = "cube_" + str(cube_side) + "_" + str(cube_side) + "_" + str(cube_side) + "_gz"
filename += ".h5pyd_test.hdfgroup.org"
#filename += ".h5"
print("filename:", filename)

f = h5pyd.File(filename, "w", username=USER_NAME, password=USER_PASSWD, endpoint=ENDPOINT)
# f = h5pyd.File(filename, "w")

print("create dataset")

dset = f.create_dataset('dset', (cube_side, cube_side, cube_side), compression="gzip", dtype='int8')

print("name:", dset.name)
print("shape:", dset.shape)
print("dset.type:", dset.dtype)
print("dset.maxshape:", dset.maxshape)

print("writing data...")

for i in range(cube_side):
    arr = np.zeros((cube_side, cube_side), dtype=dset.dtype)
    mid = cube_side // 2
    arr[mid,:] = 22
    arr[:,mid] = 44
    dset[i,:,:] = arr
print("done!")

f.close()





import json

# rom .config import Config

"""
get aiobotocore lambda client
"""

LAMBDA_REQ_PREFIX = "http+lambda://"

class Session:
    def __init__(self, timeout=10):
        self.timeout = timeout

    def mount(self, protocol, adapter):
        # TBD
        print(f"requests_lambda mount({protocol})")

    def _invoke(self, req, method="GET", params=None, headers=None): 
        if not req:
            msg = "no req"
            raise ValueError(msg)
        if not req.startswith(LAMBDA_REQ_PREFIX):
            msg = f"Expected req to start with {LAMBDA_REQ_PREFIX}"
            raise ValueError(msg)
        if method not in ("GET", "PUT", "POST", "DELETE"):
            msg = f"Unexpected method: {method}"
            raise ValueError(msg)

        s = req[len(LAMBDA_REQ_PREFIX):] # strip off protocol
        index = s.find('/')
        if index <= 0:
            msg = "Unexpected request"
            raise ValueError(msg)
        function_name = s[:(index-1)]
        req_path = s[:index]
        if not req_path:
            msg = "no request path found"
            raise ValueError(msg)

        req_json = {"method": method,
                   "request": req_path,
                   "params": params,
                   "headers": headers
                   }

        payload = json.dumps(req_json).encode('utf-8')

        import boto3  # import here so it's not a global dependency

        with boto3.client('lambda') as lambda_client:
            lambda_rsp = lambda_client.invoke(
                     FunctionName=function_name,
                     InvocationType='RequestResponse',
                     Payload=payload)
        print("got lambda rsp:", lambda_rsp)
        return lambda_rsp

    def get(self, req, params=None, headers=None):
        """
        Lambda GET request

        req should be in form: "http+lambda://function/path"
        """
        rsp = self._invoke(req, params=params, headers=headers)
        return rsp

    def put(self, req, params=None, headers=None):
        """
        Lambda PUT request

        req should be in form: "http+lambda://function/path"
        """
        rsp = self._invoke(req, method="PUT", params=params, headers=headers)
        return rsp

    def post(self, req, params=None, headers=None):
        """
        Lambda POST request

        req should be in form: "http+lambda://function/path"
        """
        rsp = self._invoke(req, method="POST", params=params, headers=headers)
        return rsp

    def delete(self, req, params=None, headers=None):
        """
        Lambda DELETE request

        req should be in form: "http+lambda://function/path"
        """
        rsp = self._invoke(req, method="DELETE", params=params, headers=headers)
        return rsp

    def close(self):
        # TBD - release any held resources
        pass

        






##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import os
import numpy as np
import config
import logging
if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

from common import ut, TestCase

def get_filename():
    if config.get("use_h5py"):
        dirpath = "data" 
    else:
        dirpath = config.get("H5PYD_TEST_FOLDER")
    filename = os.path.join(dirpath, "diamond.h5")
    return filename


class TestDiamondInspect(TestCase):

    def setUp(self):
        
        filename = get_filename()
        self.f = h5py.File(filename, "r")

    def test_obj_count(self):
        counts = {"groups": 0, "datasets": 0}
        def visit(name):
            obj = self.f[name]
            if isinstance(obj, h5py.Dataset):
                counts["datasets"] += 1
            elif isinstance(obj, h5py.Group):
                counts["groups"] += 1
            else:
                # unexpected class
                self.assertTrue(False)
        self.f.visit(visit)
        self.assertEqual(2, counts["groups"])
        self.assertEqual(1, counts["datasets"])

    def test_links(self):
        self.assertEqual(len(self.f), 2)
        self.assertTrue("g1" in self.f)
        self.assertTrue("g2" in self.f)
        g1 = self.f["g1"]
        self.assertEqual(len(g1), 1)
        self.assertTrue("dset" in g1)
        g1_dset = g1["dset"]
        self.assertEqual(g1_dset[...], 42)
        g2 = self.f["g2"]
        self.assertEqual(len(g2), 1)
        self.assertTrue("dset" in g2)
        g2_dset = g1["dset"]
        self.assertEqual(g1_dset.id.id, g2_dset.id.id)

if __name__ == '__main__':
    print("filename:", get_filename())
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    
    ut.main()




##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HSDS (HDF5 Scalable Data Service), Libraries and      #
# Utilities.  The full HSDS copyright notice, including                      #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import logging

try:
    import h5py
    import h5pyd
    import numpy as np
except ImportError as e:
    sys.stderr.write("ERROR : %s : install it to use this utility...\n" % str(e))
    sys.exit(1)

if __name__ == "utillib":
    from chunkiter import ChunkIterator
else:
    from .chunkiter import ChunkIterator


def dump_dtype(dt):
    if not isinstance(dt, np.dtype):
        raise TypeError("expected np.dtype, but got: {}".format(type(dt)))
    if len(dt) > 0:
        out = "{"
        for name in dt.fields:
            subdt = dt.fields[name][0]
            out += "{}: {} |".format(name, dump_dtype(subdt))
        out = out[:-1] + "}"
    else:
        ref = h5py.check_dtype(ref=dt)
        if ref:
            out = str(ref)
        else:
            vlen = h5py.check_dtype(vlen=dt)
            if vlen:
                out = "VLEN: " + dump_dtype(vlen)
            else:
                out = str(dt)
    return out


def is_h5py(obj):
    # Return True if objref is a h5py object and False is not
    if isinstance(obj, object) and isinstance(obj.id.id, int):
        return True
    else:
        return False


def is_reference(val):
    try:
        if isinstance(val, object) and val.__class__.__name__ == "Reference":
            return True
        elif isinstance(val, type) and val.__name__ == "Reference":
            return True
    except AttributeError as ae:
        msg = "is_reference for {} error: {}".format(val, ae)
        logging.error(msg)

    return False


def is_regionreference(val):
    try:
        if isinstance(val, object) and val.__class__.__name__ == "RegionReference":
            return True
        elif isinstance(val, type) and val.__name__ == "RegionReference":
            return True
    except AttributeError as ae:
        msg = "is_reference for {} error: {}".format(val, ae)
        logging.error(msg)

    return False


def has_reference(dtype):
    has_ref = False
    if not isinstance(dtype, np.dtype):
        return False
    if len(dtype) > 0:
        for name in dtype.fields:
            item = dtype.fields[name]
            if has_reference(item[0]):
                has_ref = True
                break
    elif dtype.metadata and 'ref' in dtype.metadata:
        basedt = dtype.metadata['ref']
        has_ref = is_reference(basedt)
    elif dtype.metadata and 'vlen' in dtype.metadata:
        basedt = dtype.metadata['vlen']
        has_ref = has_reference(basedt)
    return has_ref

def is_vlen(dtype):
    if dtype.metadata and 'vlen' in dtype.metadata:
        return True
    else:
        return False

def get_fillvalue(dset):

    fillvalue = None
    if not is_vlen(dset.dtype):
        try:
            # can trigger a runtime error if fillvalue is undefined
            fillvalue = dset.fillvalue
        except RuntimeError:
            pass  # ignore
    return fillvalue

def is_compact(dset):
    if isinstance(dset.id.id, str):
        return False  # compact storage not used with HSDS
    cpl = dset.id.get_create_plist()
    if cpl.get_layout() == h5py.h5d.COMPACT:
        return True
    else:
        return False

def convert_dtype(srcdt, ctx):
    """ Return a dtype based on input dtype, converting any Reference types from
    h5py style to h5pyd and vice-versa.
    """

    msg = "convert dtype: {}, type: {},".format(srcdt, type(srcdt))
    logging.info(msg)

    if len(srcdt) > 0:
        fields = []
        for name in srcdt.fields:
            item = srcdt.fields[name]
            # item is a tuple of dtype and integer offset
            field_dt = convert_dtype(item[0], ctx)
            fields.append((name, field_dt))
        tgt_dt = np.dtype(fields)
    else:
        # check if this a "special dtype"
        if srcdt.metadata and 'ref' in srcdt.metadata:
            ref = srcdt.metadata['ref']
            if is_reference(ref):
                if is_h5py(ctx['fout']):
                    tgt_dt = h5py.special_dtype(ref=h5py.Reference)
                else:
                    tgt_dt = h5pyd.special_dtype(ref=h5pyd.Reference)
            elif is_regionreference(ref):
                if is_h5py(ctx['fout']):
                    tgt_dt = h5py.special_dtype(ref=h5py.RegionReference)
                else:
                    tgt_dt = h5py.special_dtype(ref=h5py.RegionReference)
            else:
                msg = "Unexpected ref type: {}".format(srcdt)
                logging.error(msg)
                raise TypeError(msg)
        elif srcdt.metadata and 'vlen' in srcdt.metadata:
            src_vlen = srcdt.metadata['vlen']
            if isinstance(src_vlen, np.dtype):
                tgt_base = convert_dtype(src_vlen, ctx)
            else:
                tgt_base = src_vlen
            if is_h5py(ctx['fout']):
                tgt_dt = h5py.special_dtype(vlen=tgt_base)
            else:
                tgt_dt = h5pyd.special_dtype(vlen=tgt_base)
        else:
            tgt_dt = srcdt
    return tgt_dt


#----------------------------------------------------------------------------------
def copy_element(val, src_dt, tgt_dt, ctx):
    logging.debug("copy_element, val: " + str(val) + " val type: " + str(type(val)) + "src_dt: " + dump_dtype(src_dt) + " tgt_dt: " + dump_dtype(tgt_dt))

    fin = ctx["fin"]
    fout = ctx["fout"]
    out = None
    if len(src_dt) > 0:
        out_fields = []
        i = 0
        for name in src_dt.fields:
            field_src_dt = src_dt.fields[name][0]
            field_tgt_dt = tgt_dt.fields[name][0]
            field_val = val[i]
            i += 1
            out_field = copy_element(field_val, field_src_dt, field_tgt_dt, ctx)
            out_fields.append(out_field)
            out = tuple(out_fields)
    elif src_dt.metadata and 'ref' in src_dt.metadata:
        if not tgt_dt.metadata or 'ref' not in tgt_dt.metadata:
            raise TypeError("Expected tgt dtype to be ref, but got: {}".format(tgt_dt))
        ref = tgt_dt.metadata['ref']
        if is_reference(ref):
            # initialize out to null ref
            if is_h5py(ctx['fout']):
                out = h5py.Reference()  # null h5py ref
            else:
                out = ''  # h5pyd refs are strings

            if ref:
                try:
                    fin_obj = fin[val]
                except AttributeError as ae:
                    msg = "Unable able to get obj for ref value: {}".format(ae)
                    logging.error(msg)
                    print(msg)
                    return None

                # TBD - for hsget, the name property is not getting set
                h5path = fin_obj.name
                if not h5path:
                    msg = "No path found for ref object"
                    logging.warn(msg)
                    if ctx["verbose"]:
                        print(msg)
                else:
                    fout_obj = fout[h5path]
                    if is_h5py(ctx['fout']):
                        out = fout_obj.ref
                    else:
                        out = str(fout_obj.ref)  # convert to string for JSON serialization

        elif is_regionreference(ref):
            out = "tbd"
        else:
            raise TypeError("Unexpected ref type: {}".format(type(ref)))
    elif src_dt.metadata and 'vlen' in src_dt.metadata:
        logging.debug("copy_elment, got vlen element, dt: {}".format(src_dt.metadata["vlen"]))
        if not isinstance(val, np.ndarray):
            raise TypeError("Expecting ndarray or vlen element, but got: {}".format(type(val)))
        if not tgt_dt.metadata or 'vlen' not in tgt_dt.metadata:
            raise TypeError("Expected tgt dtype to be vlen, but got: {}".format(tgt_dt))
        src_vlen_dt = src_dt.metadata["vlen"]
        tgt_vlen_dt = tgt_dt.metadata["vlen"]
        if has_reference(src_vlen_dt):
            if len(val.shape) == 0:
                # scalar array
                e = val[()]
                v = copy_element(e, src_vlen_dt, tgt_vlen_dt, ctx)
                out = np.array(v, dtype=tgt_dt)
            else:

                out = np.zeros(val.shape, dtype=tgt_dt)
                for i in range(len(out)):
                    e = val[i]
                    out[i] = copy_element(e, src_vlen_dt, tgt_vlen_dt, ctx)
        else:
            # can just directly copy the array
            out = np.zeros(val.shape, dtype=tgt_dt)
            out[...] = val[...]
    else:
        out = val  # can just copy as is
    return out


#----------------------------------------------------------------------------------
def copy_array(src_arr, ctx):
    """ Copy the numpy array to a new array.
    Convert any reference type to point to item in the target's hierarchy.
    """
    if not isinstance(src_arr, np.ndarray):
        raise TypeError("Expecting ndarray, but got: {}".format(src_arr))
    tgt_dt = convert_dtype(src_arr.dtype, ctx)
    tgt_arr = np.zeros(src_arr.shape, dtype=tgt_dt)

    if has_reference(src_arr.dtype):
        # flatten array to simplify iteration
        count = np.product(src_arr.shape)
        tgt_arr_flat = tgt_arr.reshape((count,))
        src_arr_flat = src_arr.reshape((count,))
        for i in range(count):
            e = src_arr_flat[i]
            element = copy_element(e, src_arr.dtype, tgt_dt, ctx)
            tgt_arr_flat[i] = element
        tgt_arr = tgt_arr_flat.reshape(src_arr.shape)
    else:
        # can just copy the entire array
        tgt_arr[...] = src_arr[...]
    return tgt_arr


#----------------------------------------------------------------------------------
def copy_attribute(desobj, name, srcobj, ctx):

    msg = "creating attribute {} in {}".format(name, srcobj.name)
    logging.debug(msg)
    if ctx["verbose"]:
        print(msg)

    tgtarr = None
    data = srcobj.attrs[name]
    src_dt = None
    try:
        src_dt = data.dtype
    except AttributeError:
        pass  # auto convert to numpy array
    # First, make sure we have a NumPy array.
    if is_h5py(srcobj):
        src_empty = h5py.Empty
    else:
        src_empty = h5pyd.Empty
    if is_h5py(desobj):
        des_empty = h5py.Empty
    else:
        des_empty = h5pyd.Empty
    
    if isinstance(data, src_empty):
        # create Empty object with tgt dtype
        tgt_dt = convert_dtype(src_dt, ctx)
        tgtarr = des_empty(tgt_dt)
    else:
        srcarr = np.asarray(data, order='C', dtype=src_dt)
        tgtarr = copy_array(srcarr, ctx)
    try:
        desobj.attrs.create(name, tgtarr)
    except (IOError, TypeError) as e:
        msg = "ERROR: failed to create attribute {} of object {} -- {}".format(name, desobj.name, str(e))
        logging.error(msg)
        print(msg)


#----------------------------------------------------------------------------------
def create_dataset(dobj, ctx):
    """ create a dataset using the properties of the passed in h5py dataset.
        If successful, proceed to copy attributes and data.
    """
    msg = "creating dataset {}, shape: {}, type: {}".format(dobj.name, dobj.shape, dobj.dtype)
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)
    fout = ctx["fout"]

    chunks = None
    dset = None

    if ctx["dataload"] == "link" and not is_vlen(dobj.dtype) and dobj.shape is not None and not is_compact(dobj):
        dset_dims = dobj.shape
        logging.debug("dset_dims: {}".format(dset_dims))
        rank = len(dset_dims)
        chunk_dims = dobj.chunks
        logging.debug("chunk_dims: {}".format(chunk_dims))
        num_chunks = 0
        dsetid = dobj.id
        spaceid = dsetid.get_space()
        if chunk_dims:
            num_chunks = dsetid.get_num_chunks(spaceid)

        logging.debug("num_chunks: {}".format(num_chunks))

        chunks = {}  # pass a map to create_dataset

        if num_chunks == 0:
            chunks["class"] = 'H5D_CONTIGUOUS_REF'
            chunks["file_uri"] = ctx["s3path"]
            chunks["offset"] = dsetid.get_offset()
            # TBD - check the size is not too large
            chunks["size"] = dsetid.get_storage_size()
            logging.info(f"using chunk layout: {chunks}")

        elif num_chunks < 10:
            # construct map of chunks
            chunk_map = {}
            for i in range(num_chunks):
                chunk_info = dsetid.get_chunk_info(i, spaceid)
                index = chunk_info.chunk_offset
                logging.debug(f"got chunk_info: {chunk_info} for chunk: {i}")
                if not isinstance(index, tuple) or len(index) != rank:
                    msg = f"Unexpected array_offset: {index} for dataset with rank: {rank}"
                    logging.error(msg)
                    raise IOError(msg)
                chunk_key = ""
                for dim in range(rank):
                    chunk_key += str(index[dim] // chunk_dims[dim])
                    if dim < rank - 1:
                        chunk_key += "_"
                logging.debug("adding chunk_key: {}".format(chunk_key))
                chunk_map[chunk_key] = (chunk_info.byte_offset, chunk_info.size)

            chunks["class"] = 'H5D_CHUNKED_REF'
            chunks["file_uri"] = ctx["s3path"]
            chunks["dims"] = dobj.chunks
            chunks["chunks"] = chunk_map
            logging.info("using chunk layout: {}".format(chunks))

        else:
            # create anonymous dataset to hold chunk info
            dt = np.dtype([('offset', np.int64), ('size', np.int32)])

            chunkinfo_arr_dims = []
            for dim in range(rank):
                chunkinfo_arr_dims.append(int(np.ceil(dset_dims[dim] / chunk_dims[dim])))
            chunkinfo_arr_dims = tuple(chunkinfo_arr_dims)
            logging.debug("creating chunkinfo array of shape: {}".format(chunkinfo_arr_dims))
            chunkinfo_arr = np.zeros(np.prod(chunkinfo_arr_dims), dtype=dt)
            for i in range(num_chunks):
                chunk_info = dsetid.get_chunk_info(i, spaceid)
                index = chunk_info.chunk_offset
                if not isinstance(index, tuple) or len(index) != rank:
                    msg = f"Unexpected array_offset: {index} for dataset with rank: {rank}"
                    logging.error(msg)
                    raise IOError(msg)
                offset = 0
                stride = 1
                for i in range(rank):
                    dim = rank - i - 1
                    offset += (index[dim] // chunk_dims[dim]) * stride
                    stride *= chunkinfo_arr_dims[dim]
                chunkinfo_arr[offset] = (chunk_info.byte_offset, chunk_info.size)
            anon_dset = fout.create_dataset(None, shape=chunkinfo_arr_dims, dtype=dt)
            anon_dset[...] = chunkinfo_arr
            logging.debug("anon_dset: {}".format(anon_dset))
            #logging.debug("anon_values: {}".format(anon_dset[...]))
            chunks["class"] = 'H5D_CHUNKED_REF_INDIRECT'
            chunks["file_uri"] = ctx["s3path"]
            chunks["dims"] = dobj.chunks
            chunks["chunk_table"] = anon_dset.id.id
            logging.info("using chunk layout: {}".format(chunks))

    # use the source object layout if we are not using reference mapping
    if chunks is None and dobj.chunks:
        chunks = tuple(dobj.chunks)

    try:
        tgt_dtype = convert_dtype(dobj.dtype, ctx)
        if dobj.shape is None or len(dobj.shape) == 0 or (is_vlen(dobj.dtype) and is_h5py(fout)):
            # don't use compression/chunks for scalar datasets
            # or vlen
            compression = None
            compression_opts = None
            chunks = None
            shuffle = None
            fletcher32 = None
            maxshape = None
            scaleoffset = None
        else:
            compression = dobj.compression
            compression_opts = dobj.compression_opts
            if ctx["default_compression"] is not None and compression is None:
                compression = ctx["default_compression"]
                compression_opts = ctx["default_compression_opts"]
                if compression and ctx["verbose"]:
                    print("applying {} filter with level: {}".format(compression, compression_opts))
            shuffle = dobj.shuffle
            fletcher32 = dobj.fletcher32
            maxshape = dobj.maxshape
            scaleoffset = dobj.scaleoffset

        fillvalue=get_fillvalue(dobj)
        dset = fout.create_dataset(
            dobj.name, shape=dobj.shape, dtype=tgt_dtype, chunks=chunks,
            compression=compression, shuffle=shuffle, maxshape=maxshape,
            fletcher32=fletcher32, compression_opts=compression_opts,
            fillvalue=fillvalue, scaleoffset=scaleoffset)
        msg = "dataset created, uuid: {}, chunk_size: {}".format(dset.id.id, str(dset.chunks))
        msg += " chunks: {}".format(chunks)
        logging.info(msg)
        if ctx["verbose"]:
            print(msg)
        logging.debug("adding dataset id {} to {} in srcid_desobj_map".format(dobj.id.id, dset))
        srcid_desobj_map = ctx["srcid_desobj_map"]
        srcid_desobj_map[dobj.id.__hash__()] = dset
    except (IOError, TypeError, KeyError) as e:
        msg = "ERROR: failed to create dataset: {}".format(str(e))
        logging.error(msg)

    return dset
# create_dataset


#----------------------------------------------------------------------------------
def write_dataset(src, tgt, ctx):
    """ write values from src dataset to target dataset.
    """
    msg = "write_dataset src: {} to tgt: {}, shape: {}, type: {}".format(src.name, tgt.name, src.shape, src.dtype)
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)

    if src.shape is None:
        # null space dataset
        msg = "no data for null space dataset: {}".format(src.name)
        logging.info(msg)
        if ctx["verbose"]:
            print(msg)
        return  # no data

    if len(src.shape) == 0:
        # scalar dataset
        x = src[()]
        msg = "writing for scalar dataset: {}".format(src.name)
        logging.info(msg)
        if ctx["verbose"]:
            print(msg)
        tgt[()] = x
        return


    fillvalue = get_fillvalue(src)

    msg = "iterating over chunks for {}".format(src.name)
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)
    try:
        it = ChunkIterator(tgt)

        logging.debug("src dtype: {}".format(src.dtype))
        logging.debug("des dtype: {}".format(tgt.dtype))

        for s in it:
            arr = src[s]
            # don't write arr if it's all zeros (or the fillvalue if defined)
            empty_arr = np.zeros(arr.shape, dtype=arr.dtype)
            if fillvalue:
                empty_arr.fill(fillvalue)
            if np.array_equal(arr, empty_arr):
                msg = "skipping chunk for slice: {}".format(str(s))
            else:
                msg = "writing dataset data for slice: {}".format(s)
                tgt[s] = arr
            logging.info(msg)
            if ctx["verbose"]:
                print(msg)
    except (IOError, TypeError) as e:
        msg = "ERROR : failed to copy dataset data : {}".format(str(e))
        logging.error(msg)
        print(msg)
    msg = "done with dataload for {}".format(src.name)
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)
    logging.info("flush fout")

# write_dataset


def create_links(gsrc, gdes, ctx):
    # add soft and external links
    srcid_desobj_map = ctx["srcid_desobj_map"]
    if ctx["verbose"]:
        print("create_links: {}".format(gsrc.name))
    for title in gsrc:
        msg = "got link: {}".format(title)
        if ctx["verbose"]:
            print(msg)
        logging.info(msg)
        lnk = gsrc.get(title, getlink=True)
        link_classname = lnk.__class__.__name__
        if link_classname == "HardLink":
            logging.debug("Got hardlink: {} gsrc: {} gdes: {}".format(title, gsrc, gdes))
            if title not in gdes:
                msg = "creating link {} with title: {}".format(gdes, title)
                if ctx["verbose"]:
                    print(msg)
                logging.info(msg)
                src_obj_id = gsrc[title].id
                src_obj_id_hash = src_obj_id.__hash__()
                logging.debug("got src_obj_id hash: {}".format(src_obj_id_hash))
                if src_obj_id_hash in srcid_desobj_map:
                    des_obj = srcid_desobj_map[src_obj_id_hash]
                    logging.debug("creating hardlink to {}".format(des_obj.id.id))
                    gdes[title] = des_obj
                else:
                    # TBD - in hdf5 1.10 it seems that two references to the same object
                    # can return different id's.  This will cause HDF5 files with
                    # multilinks to not load correctly
                    msg = "could not find map item to src id: {}".format(src_obj_id_hash)
                    logging.warn(msg)
                    if ctx["verbose"]:
                        print("WARNING: " + msg)
        elif link_classname == "SoftLink":
            msg = "creating SoftLink({}) with title: {}".format(lnk.path, title)
            if ctx["verbose"]:
                print(msg)
            logging.info(msg)
            if is_h5py(gdes):
                soft_link = h5py.SoftLink(lnk.path)
            else:
                soft_link = h5pyd.SoftLink(lnk.path)
            gdes[title] = soft_link
        elif link_classname == "ExternalLink":
            msg = "creating ExternalLink({}, {}) with title: {}".format(lnk.filename, lnk.path, title)
            if ctx["verbose"]:
                print(msg)
            logging.info(msg)
            if is_h5py(gdes):
                ext_link = h5py.ExternalLink(lnk.filename, lnk.path)
            else:
                ext_link = h5pyd.ExternalLink(lnk.filename, lnk.path)
            gdes[title] = ext_link
        else:
            msg = "Unexpected link type: {}".format(lnk.__class__.__name__)
            logging.warning(msg)
            if ctx["verbose"]:
                print(msg)
    logging.info("flush fout")


#----------------------------------------------------------------------------------
def create_group(gobj, ctx):
    msg = "creating group {}".format(gobj.name)
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)
    fout = ctx["fout"]
    grp = fout.create_group(gobj.name)
    srcid_desobj_map = ctx["srcid_desobj_map"]
    logging.debug("adding group id {} to {} in srcid_desobj_map".format(gobj.id.id, grp))
    srcid_desobj_map[gobj.id.__hash__()] = grp

    return grp
# create_group


# -----------------------------------------------------------------------------
def create_datatype(obj, ctx):
    msg = "creating datatype {}".format(obj.name)
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)
    fout = ctx["fout"]
    fout[obj.name] = obj.dtype
    srcid_desobj_map = ctx["srcid_desobj_map"]
    logging.debug("adding datatype id {} to {} in srcid_desobj_map".format(obj.id.id, fout[obj.name]))
    srcid_desobj_map[obj.id.__hash__()] = fout[obj.name]


# create_datatype

#----------------------------------------------------------------------------------
def load_file(fin, fout, verbose=False, dataload="ingest", s3path=None, compression=None, compression_opts=None):
    logging.info("input file: {}".format(fin.filename))
    logging.info("output file: {}".format(fout.filename))
    if dataload != "ingest":
        if not dataload:
            logging.info("no data load")
        elif dataload == "link":
            if not s3path:
                logging.error("s3path expected to be set")
                sys.exit(1)
            logging.info("using s3path")
        else:
            logging.error("unexpected dataload value:", dataload)
            sys.exit(1)

    # it would be nice to make a class out of these functions, but that
    # makes it hard to use visititems iterator.
    # instead, create a context object to pass arround common state
    ctx = {}
    ctx["fin"] = fin
    ctx["fout"] = fout
    ctx["verbose"] = verbose
    ctx["dataload"] = dataload  # ingest, link, or None
    ctx["default_compression"] = compression
    ctx["default_compression_opts"] = compression_opts
    ctx["s3path"] = s3path
    ctx["srcid_desobj_map"] = {}

    def copy_attribute_helper(name, obj):
        logging.info(f"copy attribute - name: {name}  obj: {obj.name}")
        tgt = fout[name]
        for a in obj.attrs:
            copy_attribute(tgt, a, obj, ctx)

    def object_create_helper(name, obj):
        logging.info(f"object create helper - name: {name} obj: {obj.name}")
        class_name = obj.__class__.__name__
        if class_name in ("Dataset", "Table"):
            create_dataset(obj, ctx)
        elif class_name == "Group":
            create_group(obj, ctx)
        elif class_name == "Datatype":
            create_datatype(obj, ctx)

    def object_link_helper(name, obj):
        class_name = obj.__class__.__name__
        logging.info("object_link_helper for object: {}".format(obj.name))
        if class_name == "Group":
            # create any soft/external links
            fout = ctx["fout"]
            grp = fout[name]
            create_links(obj, grp, ctx)

    def object_copy_helper(name, obj):
        class_name = obj.__class__.__name__
        logging.debug("object_copy_helper for object: {}".format(obj.name))

        if class_name in ("Dataset", "Table"):
            if ctx["dataload"] == "link" and not is_vlen(obj.dtype) and not is_compact(obj) and len(obj.shape) > 0:
                logging.info("skip datacopy for link reference")
            else:
                logging.debug("calling write_dataset for dataset: {}".format(obj.name))
                tgt = fout[obj.name]
                write_dataset(obj, tgt, ctx)
        elif class_name == "Group":
            logging.debug("skip copy for group: {}".format(obj.name))
        elif class_name == "Datatype":
            logging.debug("skip copy for datatype: {}".format(obj.name))
        else:
            logging.error("no handler for object class: {}".format(type(obj)))

    # build a rough map of the file using the internal function above
    # copy over any attributes
    # create soft/external links (and hardlinks not already created)
    logging.info("creating target objects and attributes")

    # build a rough map of the file using the internal function above
    logging.info("creating target objects")
    fin.visititems(object_create_helper)

    # copy over any attributes
    logging.info("creating target attributes")
    fin.visititems(copy_attribute_helper)

    # create soft/external links (and hardlinks not already created)
    create_links(fin, fout, ctx)  # create root soft/external links
    fin.visititems(object_link_helper)

    if dataload == "ingest" or dataload == "link":
        # copy dataset data
        logging.info("copying dataset data")
        fin.visititems(object_copy_helper)
    else:
        logging.info("skipping dataset data copy (dataload is None)")

    # create any root attributes
    for ga in fin.attrs:
        copy_attribute(fout, ga, fin, ctx)

    # Fully flush the h5py handle.
    fout.close()

    # close up the source domain, see reason(s) for this below
    fin.close()
    msg = "load_file complete"
    logging.info(msg)
    if verbose:
        print(msg)

    return 0
    # load_file



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import os
import numpy as np
import config
import logging
if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

from common import ut, TestCase

def get_filename():
    if config.get("use_h5py"):
        dirpath = "data" 
    else:
        dirpath = config.get("H5PYD_TEST_FOLDER")
    filename = os.path.join(dirpath, "shuffle_compress.h5")
    return filename


class TestShuffleInspect(TestCase):

    def setUp(self):
        
        filename = get_filename()
        self.f = h5py.File(filename, "r")


    def test_dset(self):
        self.assertEqual(len(self.f), 1)
        self.assertTrue("dset" in self.f)
        dset = self.f["dset"]
        self.assertEqual(dset.compression, "gzip")
        self.assertTrue(dset.shuffle)
        self.assertEqual(dset.shape, (100,))
        arr = dset[0:10]
        for  i in range(10):
            self.assertEqual(arr[i], i)

if __name__ == '__main__':
    print("filename:", get_filename())
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    
    ut.main()




##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
import logging
import numpy as np

import config

if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

from common import ut, TestCase
from datetime import datetime


class TestCreateDataset(TestCase):


    def test_create_simple_dset(self):
        filename = self.getFileName("create_simple_dset")
        print("filename:", filename)
        f = h5py.File(filename, "w")

        dims = (40, 80)
        dset = f.create_dataset('simple_dset', dims, dtype='f4')

        self.assertEqual(dset.name, "/simple_dset")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 2)
        self.assertEqual(dset.ndim, 2)
        self.assertEqual(dset.shape[0], 40)
        self.assertEqual(dset.shape[1], 80)
        self.assertEqual(str(dset.dtype), 'float32')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 2)
        self.assertEqual(dset.maxshape[0], 40)
        self.assertEqual(dset.maxshape[1], 80)
        self.assertEqual(dset[0,0], 0)

        dset_ref = f['/simple_dset']
        self.assertTrue(dset_ref is not None)
        if not config.get("use_h5py"):
            # obj ids should be the same with h5pyd (but not h5py)
            self.assertEqual(dset.id.id, dset_ref.id.id)
            # Check dataset's last modified time
            self.assertTrue(isinstance(dset.modified, datetime))

        if h5py.__name__ == "h5pyd":
            # test h5pyd extensions
            if not config.get('use_h5py') and isinstance(f.id.id, str) and f.id.id.startswith("g-"):
                self.assertEqual(dset.num_chunks, 0)
                self.assertEqual(dset.allocated_size, 0)

        # try with chunk=True
        dset_chunked = f.create_dataset('chunked_dset', dims, dtype='f4', chunks=True)
        if config.get('use_h5py') or (isinstance(f.id.id, str) and f.id.id.startswith("g-")):
            self.assertTrue(dset_chunked.chunks)
        else:
            # h5serv not reporting chunks
            self.assertTrue(dset_chunked.chunks is None)

        f.close()

    def test_create_float16_dset(self):

        filename = self.getFileName("create_float16_dset")
        print("filename:", filename)
        f = h5py.File(filename, "w")
        if not config.get('use_h5py') and isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            # Float16 not supported with h5serv
            return

        nrows = 4
        ncols = 8
        dims = (nrows, ncols)
        dset = f.create_dataset('simple_dset', dims, dtype='f2')

        self.assertEqual(dset.name, "/simple_dset")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 2)
        self.assertEqual(dset.shape[0], nrows)
        self.assertEqual(dset.shape[1], ncols)
        self.assertEqual(str(dset.dtype), 'float16')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 2)
        self.assertEqual(dset.maxshape[0], nrows)
        self.assertEqual(dset.maxshape[1], ncols)
        self.assertEqual(dset[0,0], 0)

        arr = np.zeros((nrows,ncols), dtype="f2")
        for i in range(nrows):
            for j in range(ncols):
                val  = float(i) * 10.0 + float(j)/10.0
                arr[i,j] = val

        # write entire array to dataset
        dset[...] = arr

        arr = dset[...]  # read back
        val = arr[2,4]   # test one value
        self.assertTrue(val > 20.4 - 0.01)
        self.assertTrue(val < 20.4 + 0.01)


        f.close()

    def test_fillvalue_simple_dset(self):
        filename = self.getFileName("fillvalue_simple_dset")
        print("filename:", filename)
        f = h5py.File(filename, "w")

        dims = (10,)
        dset = f.create_dataset('fillvalue_simple_dset', dims, fillvalue=0xdeadbeef, dtype='uint32')

        self.assertEqual(dset.name, "/fillvalue_simple_dset")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 1)
        self.assertEqual(dset.shape[0], 10)
        self.assertEqual(str(dset.dtype), 'uint32')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 1)
        self.assertEqual(dset.maxshape[0], 10)
        self.assertEqual(dset.fillvalue, 0xdeadbeef)
        self.assertEqual(dset[0], 0xdeadbeef)

        f.close()

    def test_fillvalue_char_dset(self):
        filename = self.getFileName("fillvalue_char_dset")
        print("filename:", filename)
        f = h5py.File(filename, "w")
        if not config.get('use_h5py') and isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            # the following is failing on h5serv
            f.close()
            return

        dims = (6, 3)
        dtype = np.dtype("S1")
        fillvalue = b'X'
        data = [[b'a', b'', b''],
                [b'b', b'', b''],
                [b'c', b'', b''],
                [b'f', b'o', b'o'],
                [b'b', b'a', b'r'],
                [b'b', b'a', b'z']]
        dset = f.create_dataset('ds1', dims, data=data, fillvalue=fillvalue, dtype=dtype)

        self.assertEqual(dset.name, "/ds1")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 2)
        self.assertEqual(dset.shape[0], 6)
        self.assertEqual(dset.shape[1], 3)
        self.assertEqual(str(dset.dtype), '|S1')
        self.assertEqual(dset.fillvalue, b'X')
        self.assertEqual(dset[0,0], b'a')
        self.assertEqual(dset[5,2], b'z')


        f.close()


    def test_simple_1d_dset(self):
        filename = self.getFileName("simple_1d_dset")
        print("filename:", filename)
        f = h5py.File(filename, "w")

        dims = (10,)
        dset = f.create_dataset('simple_1d_dset', dims, dtype='uint32')

        self.assertEqual(dset.name, "/simple_1d_dset")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 1)
        self.assertEqual(dset.ndim, 1)
        self.assertEqual(dset.shape[0], 10)
        self.assertEqual(str(dset.dtype), 'uint32')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 1)
        self.assertEqual(dset.maxshape[0], 10)
        self.assertEqual(dset.fillvalue, 0)

        self.assertEqual(dset[0], 0)

        dset[:] = np.ones((10,), dtype='uint32')
        vals = dset[:]  # read back
        for i in range(10):
            self.assertEqual(vals[i], 1)

        # Write 2's to the first five elements
        dset[0:5] = [2,] * 5
        vals = dset[:]


        f.close()

    def test_fixed_len_str_dset(self):
        filename = self.getFileName("fixed_len_str_dset")
        print("filename:", filename)
        f = h5py.File(filename, "w")

        dims = (10,)
        dset = f.create_dataset('fixed_len_str_dset', dims, dtype='|S6')

        self.assertEqual(dset.name, "/fixed_len_str_dset")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 1)
        self.assertEqual(dset.shape[0], 10)
        self.assertEqual(str(dset.dtype), '|S6')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 1)
        self.assertEqual(dset.maxshape[0], 10)
        self.assertEqual(dset.fillvalue, b'')

        self.assertEqual(dset[0], b'')

        words = (b"one", b"two", b"three", b"four", b"five", b"six", b"seven", b"eight", b"nine", b"ten")
        dset[:] = words
        vals = dset[:]  # read back
        for i in range(10):
            self.assertEqual(vals[i], words[i])

        f.close()

    def test_create_dset_by_path(self):
        filename = self.getFileName("create_dset_by_path")
        print("filename:", filename)
        f = h5py.File(filename, "w")

        dims = (40,)
        dset = f.create_dataset('/mypath/simple_dset', dims, dtype='i8')

        self.assertEqual(dset.name, "/mypath/simple_dset")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 1)
        self.assertEqual(dset.shape[0], 40)
        self.assertEqual(str(dset.dtype), 'int64')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 1)
        self.assertEqual(dset.maxshape[0], 40)

        grp = f['/mypath']
        dset_ref = grp['simple_dset']
        self.assertTrue(dset_ref is not None)
        if not config.get("use_h5py"):
            # obj ids should be the same with h5pyd (but not h5py)
            self.assertEqual(dset.id.id, dset_ref.id.id)
            # Check dataset's last modified time
            self.assertTrue(isinstance(dset.modified, datetime))
            #self.assertEqual(dset.modified.tzname(), six.u('UTC'))

        f.close()

    def test_create_dset_gzip(self):
        filename = self.getFileName("create_dset_gzip")
        print("filename:", filename)

        f = h5py.File(filename, "w")

        dims = (40, 80)

        # create some test data
        arr = np.random.rand(dims[0], dims[1])

        dset = f.create_dataset('simple_dset_gzip', data=arr, dtype='f8',
            compression='gzip', compression_opts=9)

        self.assertEqual(dset.name, "/simple_dset_gzip")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 2)
        self.assertEqual(dset.shape[0], 40)
        self.assertEqual(dset.shape[1], 80)
        self.assertEqual(str(dset.dtype), 'float64')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 2)
        self.assertEqual(dset.maxshape[0], 40)
        self.assertEqual(dset.maxshape[1], 80)

        chunks = dset.chunks  # chunk layout auto-generated
        self.assertTrue(chunks is not None)
        self.assertEqual(len(chunks), 2)
        if isinstance(dset.id.id, str) and dset.id.id.startswith("d-"):
            # HSDS will create a different chunk layout
            self.assertEqual(chunks[0], 40)
            self.assertEqual(chunks[1], 80)
        else:
            self.assertEqual(chunks[0], 20)
            self.assertEqual(chunks[1], 40)
        if isinstance(f.id.id, str) and f.id.id.startswith("g-"):
            # h5serv not setting this
            self.assertEqual(dset.compression, 'gzip')
            self.assertEqual(dset.compression_opts, 9)
        self.assertFalse(dset.shuffle)

        dset_ref = f['/simple_dset_gzip']
        self.assertTrue(dset_ref is not None)
        if not config.get("use_h5py"):
            # obj ids should be the same with h5pyd (but not h5py)
            self.assertEqual(dset.id.id, dset_ref.id.id)
            # Check dataset's last modified time
            self.assertTrue(isinstance(dset.modified, datetime))
            #self.assertEqual(dset.modified.tzname(), six.u('UTC'))

        f.close()

    def test_create_dset_lz4(self):
        filename = self.getFileName("create_dset_lz4")
        print("filename:", filename)

        f = h5py.File(filename, "w")

        if config.get("use_h5py"):
            return # lz4 not supported with h5py

        if "lz4" not in f.compressors:
            print("lz4 not supproted")
            return

        dims = (40, 80)

        # create some test data
        arr = np.random.rand(dims[0], dims[1])

        dset = f.create_dataset('simple_dset_lz4', data=arr, dtype='i4',
            compression='lz4', compression_opts=5)

        self.assertEqual(dset.name, "/simple_dset_lz4")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 2)
        self.assertEqual(dset.shape[0], 40)
        self.assertEqual(dset.shape[1], 80)
        self.assertEqual(str(dset.dtype), 'int32')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 2)
        self.assertEqual(dset.maxshape[0], 40)
        self.assertEqual(dset.maxshape[1], 80)

        chunks = dset.chunks  # chunk layout auto-generated
        self.assertTrue(chunks is not None)
        self.assertEqual(len(chunks), 2)
        if isinstance(dset.id.id, str) and dset.id.id.startswith("d-"):
            # HSDS will create a different chunk layout
            self.assertEqual(chunks[0], 40)
            self.assertEqual(chunks[1], 80)
        else:
            self.assertEqual(chunks[0], 20)
            self.assertEqual(chunks[1], 40)
        self.assertEqual(dset.compression, 'lz4')
        self.assertEqual(dset.compression_opts, 5)
        self.assertFalse(dset.shuffle)

        dset_ref = f['/simple_dset_lz4']
        self.assertTrue(dset_ref is not None)
        if not config.get("use_h5py"):
            # obj ids should be the same with h5pyd (but not h5py)
            self.assertEqual(dset.id.id, dset_ref.id.id)
            # Check dataset's last modified time
            self.assertTrue(isinstance(dset.modified, datetime))
            #self.assertEqual(dset.modified.tzname(), six.u('UTC'))

        f.close()

    def test_create_dset_gzip_and_shuffle(self):
        filename = self.getFileName("create_dset_gzip_and_shuffle")
        print("filename:", filename)

        f = h5py.File(filename, "w")

        dims = (40, 80)

        # create some test data
        arr = np.random.rand(dims[0], dims[1])
        kwds = {"chunks": (4,8)}
        dset = f.create_dataset('simple_dset_gzip_shuffle', data=arr, dtype='f8',
            compression='gzip', shuffle=True, compression_opts=9, **kwds)

        self.assertEqual(dset.name, "/simple_dset_gzip_shuffle")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 2)
        self.assertEqual(dset.shape[0], 40)
        self.assertEqual(dset.shape[1], 80)
        self.assertEqual(str(dset.dtype), 'float64')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 2)
        self.assertEqual(dset.maxshape[0], 40)
        self.assertEqual(dset.maxshape[1], 80)

        chunks = dset.chunks  # chunk layout auto-generated
        self.assertTrue(isinstance(chunks, tuple))
        self.assertEqual(len(chunks), 2)
        #self.assertEqual(dset.compression, 'gzip')
        #self.assertEqual(dset.compression_opts, 9)
        self.assertTrue(dset.shuffle)

        dset_ref = f['/simple_dset_gzip_shuffle']
        self.assertTrue(dset_ref is not None)
        if not config.get("use_h5py"):
            # obj ids should be the same with h5pyd (but not h5py)
            self.assertEqual(dset.id.id, dset_ref.id.id)
            # Check dataset's last modified time
            self.assertTrue(isinstance(dset.modified, datetime))
            #self.assertEqual(dset.modified.tzname(), six.u('UTC'))

        f.close()

    def test_bool_dset(self):
        filename = self.getFileName("bool_dset")
        print("filename:", filename)
        f = h5py.File(filename, "w")

        dims = (10,)
        dset = f.create_dataset('bool_dset', dims, dtype=bool)

        self.assertEqual(dset.name, "/bool_dset")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 1)
        self.assertEqual(dset.shape[0], 10)
        self.assertEqual(str(dset.dtype), 'bool')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 1)
        self.assertEqual(dset.maxshape[0], 10)
        self.assertEqual(dset.fillvalue, 0)

        self.assertEqual(dset[0], False)


        vals = dset[:]  # read back
        for i in range(10):
            self.assertEqual(vals[i], False)

        # Write True's to the first five elements
        dset[0:5] = [True,]*5

        dset = None
        dset = f["/bool_dset"]

        # read back
        vals = dset[...]
        for i in range(5):
            if i<5:
                self.assertEqual(vals[i], True)
            else:
                self.assertEqual(vals[i], False)

        f.close()

    def test_require_dset(self):
        filename = self.getFileName("require_dset")
        print("filename:", filename)
        f = h5py.File(filename, "w")

        self.assertEqual(len(f), 0)

        dims = (40, 80)
        dset = f.require_dataset('dset', dims, dtype='f8')

        self.assertEqual(dset.name, "/dset")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 2)
        self.assertEqual(dset.ndim, 2)
        self.assertEqual(dset.shape[0], 40)
        self.assertEqual(dset.shape[1], 80)
        self.assertEqual(str(dset.dtype), 'float64')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 2)
        self.assertEqual(dset.maxshape[0], 40)
        self.assertEqual(dset.maxshape[1], 80)
        self.assertEqual(dset[0,0], 0)

        self.assertEqual(len(f), 1)

        dset_2 = f.require_dataset('dset', dims, dtype='f8')
        if not config.get("use_h5py"):
            self.assertEqual(dset.id.id, dset_2.id.id)
        self.assertEqual(len(f), 1)

        dset_3 = f.require_dataset('dset', dims, dtype='f4')
        if not config.get("use_h5py"):
            self.assertEqual(dset.id.id, dset_3.id.id)
        self.assertEqual(str(dset_3.dtype), 'float64')

        self.assertEqual(len(f), 1)

        try: 
            f.require_dataset('dset', dims, dtype='f4', exact=True)
            self.assertTrue(False)  # exception expected
        except TypeError:
            pass

        self.assertEqual(len(f), 1)

        f.close()

    def test_create_dset_like(self):
        filename = self.getFileName("create_dset_like")
        print("filename:", filename)
        f = h5py.File(filename, "w")

        def check_props(dset):
            self.assertTrue(isinstance(dset.shape, tuple))
            self.assertEqual(len(dset.shape), 2)
            self.assertEqual(dset.ndim, 2)
            self.assertEqual(dset.shape[0], 40)
            self.assertEqual(dset.shape[1], 80)
            self.assertEqual(str(dset.dtype), 'float32')
            self.assertTrue(isinstance(dset.maxshape, tuple))
            self.assertEqual(len(dset.maxshape), 2)
            self.assertEqual(dset.maxshape[0], 40)
            self.assertEqual(dset.maxshape[1], 80)
            self.assertEqual(dset[0,0], 0)

        dims = (40, 80)
        dset = f.create_dataset('simple_dset', dims, dtype='f4')

        self.assertEqual(dset.name, '/simple_dset')
        check_props(dset)
        
        dset_copy = f.create_dataset_like('similar_dset', dset)
        self.assertEqual(dset_copy.name, '/similar_dset')
        check_props(dset_copy)

        self.assertEqual(len(f), 2)

        f.close()

    def test_create_dset_empty(self):
        filename = self.getFileName("create_dset_empty")
        print("filename:", filename)
        f = h5py.File(filename, "w")

        def check_props(dset):
            self.assertEqual(dset.shape, None)
            self.assertEqual(str(dset.dtype), 'float32')

        # create by providing a type
        dset1 = f.create_dataset('dset1', dtype='f4')
        self.assertEqual(dset1.name, '/dset1')
        check_props(dset1)

        # create using the Empty object
        dset2 = f.create_dataset('dset2', data=h5py.Empty("float32"))
        self.assertEqual(dset2.name, '/dset2')
        check_props(dset2)

         
        f.close()
        
if __name__ == '__main__':
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    ut.main()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HSDS (HDF5 Scalable Data Service), Libraries and      #
# Utilities.  The full HSDS copyright notice, including                      #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import logging
import numpy as np

try:
    import h5py
    import h5pyd
except ImportError as e:
    sys.stderr.write("ERROR : %s : install it to use this utility...\n" % str(e))
    sys.exit(1)

try:
    import s3fs
    S3FS_IMPORT = True
except ImportError:
    S3FS_IMPORT = False

if __name__ == "__main__":
    from config import Config
    from chunkiter import ChunkIterator
else:
    from .config import Config
    from .chunkiter import ChunkIterator

cfg = Config()

def diff_attrs(src, tgt, ctx):
    """ compare attributes of src and tgt """
    msg = "checking attributes of {}".format(src.name)
    logging.debug(msg)

    if len(src.attrs) != len(tgt.attrs):
        msg = "<{}> have a different number of attribute from <{}>".format(src.name, tgt.name)
        logging.info(msg)
        if not ctx["quiet"]:
            print(msg)
        ctx["differences"] += 1
        return False

    for name in src.attrs:
        msg = "checking attribute {} of {}".format(name, src.name)
        logging.debug(msg)
        if ctx["verbose"]:
            print(msg)
        if name not in tgt.attrs:
            msg = "<{}>  has attribute {} not found in <{}>".format(src.name, name, tgt.name)
            logging.info(msg)
            if not ctx["quiet"]:
                print(msg)
            ctx["differences"] += 1
            return False
        src_attr = src.attrs[name]
        tgt_attr = tgt.attrs[name]
        if isinstance(src_attr, np.ndarray):
            # compare shape, type, and values
            if src_attr.dtype != tgt_attr.dtype:
                msg = "Type of attribute {} of <{}> is different".format(name, src.name)
                logging.info(msg)
                if not ctx["quiet"]:
                    print(msg)
                ctx["differences"] += 1
                return False
            if src_attr.shape != tgt_attr.shape:
                msg = "Shape of attribute {} of <{}> is different".format(name, src.name)
                logging.info(msg)
                if not ctx["quiet"]:
                    print(msg)
                ctx["differences"] += 1
                return False
            if hash(src_attr.tostring()) != hash(tgt_attr.tostring()):
                msg = "values for attribute {} of <{}> differ".format(name, src.name)
                logging.info(msg)
                if not ctx["quiet"]:
                    print(msg)
                ctx["differences"] += 1
                return False
        elif src_attr != tgt_attr:
            # returned as int or string, just compare values
            msg = "<{}>  has attribute {} different than <{}>".format(src.name, name, tgt.name)
            logging.info(msg)

            if not ctx["quiet"]:
                print(msg)
            ctx["differences"] += 1
            return False

    # of of attribute iteration
    return True


def diff_group(src, ctx):
    """ compare group in src and tgt
    """
    msg = "checking group <{}>".format(src.name)
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)

    fout = ctx["fout"]

    if src.name not in fout:
        msg = "<{}> not found in target".format(src.name)
        logging.info(msg)
        if not ctx["quiet"]:
            print(msg)
        ctx["differences"] += 1
        return False

    tgt = fout[src.name]

    # printed when there is a difference
    output = "group: <{}> and <{}>".format(src.name, tgt.name)
    if len(src) != len(tgt):
        msg = "{} group have a different number of links from {}".format(src.name, tgt.name)
        logging.info(msg)
        if ctx["verbose"]:
            print(msg)
        if not ctx["quiet"]:
            print(output)
        ctx["differences"] += 1
        return False

    for title in src:
        if ctx["verbose"]:
            print("got link: '{}' of group <{}>".format(title, src.name))
        if title not in tgt:
            msg = "<{}> group has link {} not found in <{}>".format(src.name, title, tgt.name)
            logging.info(msg)
            if ctx["verbose"]:
                print(msg)
            if not ctx["quiet"]:
                print(output)
            ctx["differences"] += 1
            return False

        lnk_src = src.get(title, getlink=True)
        lnk_src_type = lnk_src.__class__.__name__
        lnk_tgt = tgt.get(title, getlink=True)
        lnk_tgt_type = lnk_tgt.__class__.__name__
        if lnk_src_type != lnk_tgt_type:
            msg = "<{}> group has link {} of different type than found in <{}>".format(src.name, title, tgt.name)
            logging.info(msg)
            if ctx["verbose"]:
                print(msg)
            if not ctx["quiet"]:
                print(output)
            ctx["differences"] += 1
            return False

        if lnk_src_type == "HardLink":
            logging.debug("Got hardlink: {}".format(title))
            # TBD: handle the case where multiple hardlinks point to same object
        elif lnk_src_type == "SoftLink":
            msg = "Got SoftLink({}) with title: {}".format(lnk_src.path, title)
            if ctx["verbose"]:
                print(msg)
            logging.info(msg)
            if lnk_src.path != lnk_tgt.path:
                msg = "<{}> group has link {} with different path than <{}>".format(src.name, title, tgt.name)
                if ctx["verbose"]:
                    print(msg)
                if not ctx["quiet"]:
                    print(output)
                ctx["differences"] += 1
                return False
        elif lnk_src_type == "ExternalLink":
            msg = "<{}> group has ExternalLink {} ({}, {})".format(src.name, title, lnk_src.filename, lnk_src.path)
            if ctx["verbose"]:
                print(msg)
            logging.info(msg)
            if lnk_src.filename != lnk_tgt.filename:
                msg = "<{}> group has external link {} with different filename than <{}>".format(src.name, title, tgt.name)
                if ctx["verbose"]:
                    print(msg)
                if not ctx["quiet"]:
                    print(output)
                ctx["differences"] += 1
                return False
            if lnk_src.path != lnk_tgt.path:
                msg = "<{}> group has external link {} with different path than <{}>".format(src.name, title, tgt.name)
                if ctx["verbose"]:
                    print(msg)
                if not ctx["quiet"]:
                    print(output)
                ctx["differences"] += 1
                return False
        else:
            msg = "Unexpected link type: {}".format(lnk_src_type)
            logging.warning(msg)
            if ctx["verbose"]:
                print(msg)
    # end link iteration

    if not ctx["noattr"]:
        result = diff_attrs(src, tgt, ctx)
    else:
        result = True
    return result



def diff_datatype(src, ctx):
    """ compare datatype objects in src and tgt
    """
    msg = "checking datatype <{}>".format(src.name)
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)

    fout = ctx["fout"]

    if src.name not in fout:
        msg = "<{}> not found in target".format(src.name)
        logging.info(msg)
        if not ctx["quiet"]:
            print(msg)
        ctx["differences"] += 1
        return False
    tgt = fout[src.name]

    if tgt.dtype != src.dtype:
        msg = "Type of <{}> is different".format(src.name)
        logging.info(msg)
        if not ctx["quiet"]:
            print(msg)
        ctx["differences"] += 1
        return False

    if not ctx["noattr"]:
        result = diff_attrs(src, tgt, ctx)
    else:
        result = True
    return result


def diff_dataset(src, ctx):
    """ compare dataset in src and tgt
    """
    msg = "checking dataset <{}>".format(src.name)
    logging.info(msg)
    if ctx["verbose"]:
        print(msg)

    fout = ctx["fout"]

    if src.name not in fout:
        msg = "<{}> not found in target".format(src.name)
        logging.info(msg)
        if not ctx["quiet"]:
            print(msg)
        ctx["differences"] += 1
        return False
    tgt = fout[src.name]

    try:
        tgt_shape = tgt.shape
    except AttributeError:
        msg = "<{}> in target not a dataset".format(src.name)
        logging.info(msg)
        if not ctx["quiet"]:
            print(msg)
        ctx["differences"] += 1
        return False

    # printed when there is a difference
    output = "dataset: <{}> and <{}>".format(src.name, tgt.name)
    if tgt_shape != src.shape:
        msg = "Shape of <{}> is different".format(src.name)
        logging.info(msg)
        if not ctx["quiet"]:
            print(output)
        ctx["differences"] += 1

        return False

    if tgt.dtype != src.dtype:
        msg = "Type of <{}> is different".format(src.name)
        logging.info(msg)
        if not ctx["quiet"]:
            print(output)
        ctx["differences"] += 1
        return False

    # TBD - check fillvalue

    if ctx["nodata"]:
        # skip data compare
        return True

    try:
        it = ChunkIterator(src)

        for s in it:
            msg = "checking dataset data for slice: {}".format(s)
            logging.debug(msg)

            arr_src = src[s]
            msg = "got src array {}".format(arr_src.shape)
            logging.debug(msg)
            arr_tgt = tgt[s]
            msg = "got tgt array {}".format(arr_tgt.shape)
            logging.debug(msg)

            if hash(arr_src.tostring()) != hash(arr_tgt.tostring()):
                msg = "values for dataset {} differ for slice: {}".format(src.name, s)
                logging.info(msg)
                if not ctx["quiet"]:
                    print(output)
                ctx["differences"] += 1
                return False

    except (IOError, TypeError) as e:
        msg = "ERROR : failed to copy dataset data : {}".format(str(e))
        logging.error(msg)
        print(msg)

    if not ctx["noattr"]:
        result = diff_attrs(src, tgt, ctx)
    else:
        result = True
    return result



def diff_file(fin, fout, verbose=False, nodata=False, noattr=False, quiet=False):
    ctx = {}
    ctx["fin"] = fin
    ctx["fout"] = fout
    ctx["verbose"] = verbose
    ctx["nodata"] = nodata
    ctx["noattr"] = noattr
    ctx["quiet"] = quiet
    ctx["differences"] = 0


    def object_diff_helper(name, obj):
        class_name = obj.__class__.__name__

        if class_name in ("Dataset", "Table"):
            diff_dataset(obj, ctx)
        elif class_name == "Group":
            diff_group(obj, ctx)
        elif class_name == "Datatype":
            diff_datatype(obj, ctx)
        else:
            logging.error("no handler for object class: {}".format(type(obj)))

    # check links in root group
    diff_group(fin, ctx)

    # build a rough map of the file using the internal function above
    fin.visititems(object_diff_helper)
    return ctx["differences"]


#----------------------------------------------------------------------------------
def usage():
    print("Usage:\n")
    print(("    {} [ OPTIONS ]  file  domain".format(cfg["cmd"])))
    print("")
    print("Description:")
    print("    Diff HDF5 file with domain")
    print("       file: HDF5 file ")
    print("       domain: domain")
    print("")
    print("Options:")
    print("     -v | --verbose :: verbose output")
    print("     -e | --endpoint <domain> :: The HDF Server endpoint, e.g. http://hsdshdflab.hdfgroup.org")
    print("     -u | --user <username>   :: User name credential")
    print("     -p | --password <password> :: Password credential")
    print("     -c | --conf <file.cnf>  :: A credential and config file")
    print("     --cnf-eg        :: Print a config file and then exit")
    print("     --logfile <logfile> :: logfile path")
    print("     --loglevel debug|info|warning|error :: Change log level")
    print("     --bucket <bucket_name> :: Storage bucket")
    print("     --nodata :: Do not compare dataset data")
    print("     --noattr :: Do not compare attributes")
    print("     --quiet :: Do not produce output")
    print("     -h | --help    :: This message.")
    print("")
#end print_usage


#----------------------------------------------------------------------------------
def print_config_example():
    print("# default")
    print("hs_username = <username>")
    print("hs_password = <passwd>")
    print("hs_endpoint = http://hsdshdflab.hdfgroup.org")
#print_config_example

#----------------------------------------------------------------------------------
def main():

    loglevel = logging.ERROR
    verbose = False
    nodata = False
    noattr = False
    quiet = False
    cfg["cmd"] = sys.argv[0].split('/')[-1]
    if cfg["cmd"].endswith(".py"):
        cfg["cmd"] = "python " + cfg["cmd"]
    cfg["logfname"] = None
    logfname=None
    rc = 0
    s3 = None  # s3fs instance

    src_files = []
    argn = 1
    while argn < len(sys.argv):
        arg = sys.argv[argn]
        val = None

        if arg[0] == '-' and len(src_files) > 0:
            # options must be placed before filenames
            print("options must precead source files")
            usage()
            sys.exit(-1)
        if len(sys.argv) > argn + 1:
            val = sys.argv[argn+1]
        if arg in ("-v", "--verbose"):
            verbose = True
            argn += 1
        elif arg == "--nodata":
            nodata = True
            argn += 1
        elif arg == "--noattr":
            noattr = True
            argn += 1
        elif arg in ("-q", "--quiet"):
            quiet = True
            argn += 1
        elif arg == "--loglevel":
            if val == "debug":
                loglevel = logging.DEBUG
            elif val == "info":
                loglevel = logging.INFO
            elif val == "warning":
                loglevel = logging.WARNING
            elif val == "error":
                loglevel = logging.ERROR
            else:
                print("unknown loglevel")
                usage()
                sys.exit(-1)
            argn += 2
        elif arg == '--logfile':
            logfname = val
            argn += 2
        elif arg in ("-b", "--bucket"):
            cfg["hs_bucket"] = val
            argn += 2
        elif arg in ("-h", "--help"):
            usage()
            sys.exit(0)
        elif arg in ("-e", "--endpoint"):
            cfg["hs_endpoint"] = val
            argn += 2
        elif arg in ("-u", "--username"):
            cfg["hs_username"] = val
            argn += 2
        elif arg in ("-p", "--password"):
            cfg["hs_password"] = val
            argn += 2
        elif arg == '--cnf-eg':
            print_config_example()
            sys.exit(0)
        elif arg[0] == '-':
            usage()
            sys.exit(-1)
        else:
            src_files.append(arg)
            argn += 1

    # setup logging
    logging.basicConfig(filename=logfname, format='%(levelname)s %(asctime)s %(filename)s:%(lineno)d %(message)s', level=loglevel)
    logging.debug("set log_level to {}".format(loglevel))

    # end arg parsing
    logging.info("username: {}".format(cfg["hs_username"]))
    logging.info("endpoint: {}".format(cfg["hs_endpoint"]))
    logging.info("verbose: {}".format(verbose))

    if len(src_files) < 2:
        # need at least a src and destination
        usage()
        sys.exit(-1)
    file_path = src_files[0]
    domain_path = src_files[1]

    logging.info("file: {}".format(file_path))
    logging.info("domain: {}".format(domain_path))
    if domain_path[0] != '/' or domain_path[-1] == '/':
        print("domain must be an absolute path, non-folder domain")
        usage()
        sys.exit(-1)


    if cfg["hs_endpoint"] is None:
        logging.error('No endpoint given, try -h for help\n')
        sys.exit(1)
    logging.info("endpoint: {}".format(cfg["hs_endpoint"]))

    try:

        # get a handle to input file
        if file_path.startswith("s3://"):
            if not S3FS_IMPORT:
                sys.stderr.write("Install S3FS package to load s3 files")
                sys.exit(1)

            if not s3:
                s3 = s3fs.S3FileSystem(use_ssl=False)
            try:
                fin = h5py.File(s3.open(file_path, "rb"), mode="r")
            except IOError as ioe:
                logging.error("Error opening file {}: {}".format(file_path, ioe))
                sys.exit(1)
        else:
            # regular h5py open
            try:
                fin = h5py.File(file_path, mode='r')
            except IOError as ioe:
                logging.error("Error opening file {}: {}".format(domain_path, ioe))
                sys.exit(1)

        # get the  domain
        try:
            username = cfg["hs_username"]
            password = cfg["hs_password"]
            endpoint = cfg["hs_endpoint"]
            bucket = cfg["hs_bucket"]
            fout = h5pyd.File(domain_path, 'r', endpoint=endpoint, username=username, password=password, bucket=bucket)
        except IOError as ioe:
            if ioe.errno == 404:
                logging.error("domain: {} not found".format(domain_path))
            if ioe.errno == 403:
                logging.error("No read access to domain: {}".format(domain_path))
            else:
                logging.error("Error opening file {}: {}".format(domain_path, ioe))
            sys.exit(1)


        # do the actual load
        if quiet:
            verbose = False
        rc = diff_file(fin, fout, verbose=verbose, nodata=nodata, noattr=noattr, quiet=quiet)

        if not quiet and rc > 0:
            print("{} differences found".format(rc))

        logging.info("diff_file done")

    except KeyboardInterrupt:
        logging.error('Aborted by user via keyboard interrupt.')
        sys.exit(1)

    sys.exit(rc)


# __main__
if __name__ == "__main__":
    main()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import config
import os

if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

from common import ut, TestCase
from datetime import datetime
from copy import copy
import six
import time
import logging

class TestFile(TestCase):

    def test_version(self):
        version = h5py.version.version
        # should be of form "n.n.n"
        n = version.find(".")
        self.assertTrue(n>=1)
        m = version[(n+1):].find('.')
        self.assertTrue(m>=1)

    def test_serverinfo(self):
        if h5py.__name__ == "h5pyd":
            info = h5py.getServerInfo()
            self.assertTrue("greeting" in info)
            self.assertTrue("name" in info)
            self.assertTrue("about" in info)
            self.assertTrue("endpoint" in info)
            self.assertTrue("username" in info)
            self.assertTrue("password" in info)
            if "hsds_version" in info:
                # server is HSDS
                self.assertTrue("node_count" in info)
                node_count = info["node_count"]
                self.assertTrue(node_count >= 1)
                self.assertTrue("isadmin" in info)


    def test_create(self):
        filename = self.getFileName("new_file")
        print("filename:", filename)
        now = time.time()
        f = h5py.File(filename, 'w')
        self.assertEqual(f.filename, filename)
        self.assertEqual(f.name, "/")
        self.assertTrue(f.id.id is not None)
        self.assertEqual(len(f.keys()), 0)
        self.assertEqual(f.mode, 'r+')
        self.assertTrue(h5py.is_hdf5(filename))
        is_hsds = False
        if isinstance(f.id.id, str) and f.id.id.startswith("g-"):
            is_hsds = True  # HSDS has different permission defaults
        if h5py.__name__ == "h5pyd":
            self.assertTrue(f.id.http_conn.endpoint.startswith("http"))
        self.assertTrue(f.id.id is not None)
        self.assertTrue('/' in f)
        # should not see id as a file
        self.assertFalse(h5py.is_hdf5(f.id.id))
        # Check domain's timestamps
        if h5py.__name__ == "h5pyd" and is_hsds:
            # TBD: remove is_hsds when h5serv timestamp changed to float
            #print("modified:", datetime.fromtimestamp(f.modified), f.modified)
            #print("created: ", datetime.fromtimestamp(f.created), f.created)
            #print("now:     ", datetime.fromtimestamp(now), now)
            # verify the timestamps make sense
            # we add a 30-sec margin to account for possible time skew
            # between client and server
            self.assertTrue(f.created - 30.0 < now)
            self.assertTrue(f.created + 30.0 > now)
            self.assertTrue(f.modified - 30.0 < now)
            self.assertTrue(f.modified + 30.0 > now)
            self.assertTrue(f.modified >= f.created)
        if is_hsds:
            # owner prop is just for HSDS
            self.assertTrue(len(f.owner) > 0)
            version = f.serverver
             # server version should be of form "n.n.n"
            n = version.find(".")
            self.assertTrue(n>=1)
            limits = f.limits
            for k in ('min_chunk_size', 'max_chunk_size', 'max_request_size', 'max_chunks_per_request'):
                self.assertTrue(k in limits)


        r = f['/']
        self.assertTrue(isinstance(r, h5py.Group))
        self.assertEqual(len(f.attrs.keys()), 0)

        # flush any pending changes - this would be called by f.close() internally,
        # but try here to confirm it can be called explicitly
        f.flush()

        f.close()
        self.assertEqual(f.id.id, 0)

        # re-open as read-write
        f = h5py.File(filename, 'w')
        self.assertTrue(f.id.id is not None)
        self.assertEqual(f.mode, 'r+')
        self.assertEqual(len(f.keys()), 0)
        root_grp = f['/']
        #f.create_group("subgrp")
        root_grp.create_group("subgrp")
        self.assertEqual(len(f.keys()), 1)
        f.close()
        self.assertEqual(f.id.id, 0)

        # re-open as read-only
        if is_hsds:
            wait_time = 90
            print("waiting {} seconds for root scan sync".format(wait_time))
            time.sleep(wait_time)  # let async process update obj number
        f = h5py.File(filename, 'r')
        self.assertEqual(f.filename, filename)
        self.assertEqual(f.name, "/")
        self.assertTrue(f.id.id is not None)
        self.assertEqual(len(f.keys()), 1)
        self.assertEqual(f.mode, 'r')
        self.assertTrue('/' in f)
        r = f['/']
        self.assertTrue(isinstance(r, h5py.Group))
        self.assertEqual(len(f.attrs.keys()), 0)

        # verify that trying to modify the file fails
        try:
            f.create_group("another_subgrp")
            self.assertTrue(False)  # expect exception
        except (IOError, OSError, ValueError):
            # h5py throws ValueError
            # h5pyd is throwing IOError
            pass

        try:
            f.attrs["foo"] = "bar"
            self.assertTrue(False)  # expect exception
        except (IOError, OSError):
            pass

        self.assertEqual(len(f.keys()), 1)

        if  h5py.__name__ == "h5pyd":
            # check properties that are only available for h5pyd
            # Note: num_groups won't reflect current state since the
            # data is being updated asynchronously
            if is_hsds:
                self.assertEqual(f.num_objects, 2)
                self.assertEqual(f.num_groups, 2)
            else:
                # reported as 0 for h5serv
                self.assertEqual(f.num_objects, 0)
                self.assertEqual(f.num_groups, 0)

            self.assertEqual(f.num_datasets, 0)
            self.assertEqual(f.num_datatypes, 0)
            self.assertTrue(f.allocated_bytes == 0)



        f.close()
        self.assertEqual(f.id.id, 0)

        # re-open using hdf5:// prefix
        if h5py.__name__ == "h5pyd" and is_hsds:
            if filename[0] == '/':
                filepath = "hdf5:/" + filename
            else:
                filepath = "hdf5://" + filename
            f = h5py.File(filepath, 'r')
            self.assertEqual(f.filename, filename)
            self.assertEqual(f.name, "/")
            self.assertTrue(f.id.id is not None)
            self.assertEqual(len(f.keys()), 1)
            self.assertEqual(f.mode, 'r')
            self.assertTrue('/' in f)
            f.close()


        # open in truncate mode
        f = h5py.File(filename, 'w')
        self.assertEqual(f.filename, filename)
        self.assertEqual(f.name, "/")
        self.assertTrue(f.id.id is not None)
        self.assertEqual(len(f.keys()), 0)
        self.assertEqual(f.mode, 'r+')
        self.assertTrue('/' in f)
        r = f['/']
        self.assertTrue(isinstance(r, h5py.Group))
        self.assertEqual(len(f.attrs.keys()), 0)

        f.close()
        self.assertEqual(f.id.id, 0)



    def test_open_notfound(self):
        # verify open of non-existent file throws exception

        try:
            filename = self.getFileName("no_file_here")
            print("filename:", filename)
            f = h5py.File(filename, 'r')
            self.assertTrue(False) #expect exception
        except IOError:
            pass


    def test_auth(self):
        if h5py.__name__ == "h5py":
            return  # ACLs are just for h5pyd

        filename = self.getFileName("file_auth")
        print("filename:", filename)

        f = h5py.File(filename, 'w')


        for name in ("g1", "g2", "g1/g1.1"):
            f.create_group(name)

        self.assertEqual(f.filename, filename)
        self.assertEqual(f.name, "/")
        self.assertTrue(f.id.id is not None)
        root_id = f.id.id
        self.assertEqual(len(f.keys()), 2)

        is_hsds = False
        if root_id.startswith("g-"):
            is_hsds = True  # HSDS has different permission defaults

        # no explicit ACLs yet
        file_acls = f.getACLs()
        if is_hsds:
            self.assertTrue(len(file_acls) >= 1)  # Should have at least the test_user1 acl
        else:
            self.assertEqual(len(file_acls), 0)

        if is_hsds:
            username = f.owner
        else:
            username = "test_user1"

        file_acl = f.getACL(username)
        # default owner ACL should grant full permissions
        acl_keys = ("create", "read", "update", "delete", "readACL", "updateACL")
        #self.assertEqual(file_acl["userName"], "default")
        for k in acl_keys:
            self.assertEqual(file_acl[k], True)

        # for h5serv a default acl should be available
        # hsds does not create one initially


        try:
            default_acl = f.getACL("default")
        except IOError as ioe:
            if ioe.errno == 404:
                if is_hsds:
                    pass # expected
                else:
                    self.assertTrue(False)

        if is_hsds:
            # create  public-read ACL
            default_acl = {}
            for key in acl_keys:
                if key == "read":
                    default_acl[key] = True
                else:
                    default_acl[key] = False
            default_acl["userName"] = "default"
            f.putACL(default_acl)
        f.close()


        # ooen with test_user2 should succeed for read mode
        try:
            f = h5py.File(filename, 'r', username=self.test_user2["name"], password=self.test_user2["password"])
            f.close()
        except IOError as ioe:
            self.assertTrue(False)

        # test_user2 has read access, but opening in write mode should fail
        try:
            f = h5py.File(filename, 'w', username=self.test_user2["name"], password=self.test_user2["password"])
            self.assertFalse(is_hsds)  # expect exception for hsds
        except IOError as ioe:
            self.assertTrue(is_hsds)
            self.assertEqual(ioe.errno, 403)  # user is not authorized

        # append mode w/ test_user2
        try:
            f = h5py.File(filename, 'a', username=self.test_user2["name"], password=self.test_user2["password"])
            self.assertFalse(is_hsds)  # expected exception
        except IOError as ioe:
            self.assertTrue(is_hsds)
            self.assertEqual(ioe.errno, 403)  # Forbidden


        f = h5py.File(filename, 'a')  # open for append with original username
        # add an acl for test_user2 that has only read/update access
        user2_acl = copy(default_acl)
        user2_acl["userName"] = self.test_user2["name"]
        user2_acl["read"] = True  # allow read access
        user2_acl["update"] = True
        user2_acl["readACL"] = True
        f.putACL(user2_acl)

        f.close()

        # ooen with test_user2 should succeed for read mode
        try:
            f = h5py.File(filename, 'r', username=self.test_user2["name"], password=self.test_user2["password"])
        except IOError as ioe:
            self.assertTrue(False)
        return


        # test_user2  opening in write mode should still fail
        try:
            f = h5py.File(filename, 'w', username=self.test_user2["name"], password=self.test_user2["password"])
            self.assertFalse(is_hsds)  # expected exception
        except IOError as ioe:
            self.assertTrue(is_hsds)
            self.assertEqual(ioe.errno, 403)  # user is not authorized

        # append mode w/ test_user2
        try:
            f = h5py.File(filename, 'a', username=self.test_user2["name"], password=self.test_user2["password"])
        except IOError as ioe:
            self.assertTrue(False)  # shouldn't get here

        grp = f['/']
        grp.file.close()  # try closing the file via a group reference

    def test_http_url(self):
        if h5py.__name__ == "h5pyd":
            if "HS_ENDPOINT" in os.environ:
                endpoint = os.environ["HS_ENDPOINT"]
                filename = self.getFileName("test_http_url_file")
                is_hsds = False
                f = h5py.File(filename, 'w')
                if isinstance(f.id.id, str) and f.id.id.startswith("g-"):
                    is_hsds = True  # HSDS has different permission defaults
                f.close()
                if is_hsds:
                    url = endpoint + filename
                    f = h5py.File(url, 'w')
                    self.assertEqual(f.filename, filename)
                    self.assertEqual(f.name, "/")
                    self.assertTrue(f.id.id is not None)
                    self.assertEqual(len(f.keys()), 0)
                    self.assertEqual(f.mode, 'r+')
                    f.close()
            else:
                print("set HS_ENDPOINT environment variable to enable test_http_url test")



    def test_close(self):
        filename = self.getFileName("close_file")
        print("filename:", filename)
        f = h5py.File(filename, 'w')
        self.assertTrue(f)
        f.close()
        self.assertFalse(f)


if __name__ == '__main__':
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    ut.main()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
from __future__ import absolute_import
import sys
import logging
import numpy as np
import config

if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

from common import ut, TestCase

"""
    Tests the h5py.Dataset.__getitem__ method.

    This module does not specifically test type conversion.  The "type" axis
    therefore only tests objects which interact with the slicing system in
    unreliable ways; for example, compound and array types.

    See test_dataset_getitem_types for type-conversion tests.

    Tests are organized into TestCases by dataset shape and type.  Test
    methods vary by slicing arg type.

    1. Dataset shape:
        Empty
        Scalar
        1D
        3D

    2. Type:
        Float
        Compound
        Array

    3. Slicing arg types:
        Ellipsis
        Empty tuple
        Regular slice
        Indexing
        Index list
        Boolean mask
        Field names
"""

"""
Disabled since low-level interface not supported with h5pyd
Update using new NULL dataset constructor once h5py 2.7 is out.
"""

class TestEmpty(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        filename = self.getFileName("dataset_testempty")
        print("filename:", filename)
        self.f = h5py.File(filename, 'w')
        self.dset = self.f.create_dataset('x',dtype='S10')
        self.empty_obj = h5py.Empty(np.dtype("S10"))

    def test_ndim(self):
        """ Verify number of dimensions """
        self.assertEqual(self.dset.ndim, 0)

    def test_shape(self):
        """ Verify shape """
        self.assertEqual(self.dset.shape, None)

    def test_size(self):
        """ Verify shape """
        self.assertEqual(self.dset.size, None)

    def test_nbytes(self):
        """ Verify nbytes """
        self.assertEqual(self.dset.nbytes, 0)

    def test_ellipsis(self):
        self.assertEqual(self.dset[...], self.empty_obj)

    def test_tuple(self):
        self.assertEqual(self.dset[()], self.empty_obj)

    def test_slice(self):
        """ slice -> ValueError """
        with self.assertRaises(ValueError):
            self.dset[0:4]

    def test_multi_block_slice(self):
        """ MultiBlockSlice -> ValueError """
        """ TBD """
        #with self.assertRaises(ValueError):
        #    self.dset[h5py.MultiBlockSlice()]

    def test_index(self):
        """ index -> ValueError """
        with self.assertRaises(ValueError):
            self.dset[0]

    def test_indexlist(self):
        """ index list -> ValueError """
        with self.assertRaises(ValueError):
            self.dset[[1,2,5]]

    def test_mask(self):
        """ mask -> ValueError """
        mask = np.array(True, dtype='bool')
        with self.assertRaises(ValueError):
            self.dset[mask]

    def test_fieldnames(self):
        """ field name -> ValueError """
        with self.assertRaises(ValueError):
            self.dset['field']
      

class TestScalarFloat(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        filename = self.getFileName("dataset_testscalarflot")
        print("filename:", filename)
        self.f = h5py.File(filename, 'w')
        if isinstance(self.f.id.id, str) and self.f.id.id.startswith("g-"):
            # flag if using HSDS
            self.is_hsds = True
        else:
            self.is_hsds = False
        self.data = np.array(42.5, dtype='f')
        self.dset = self.f.create_dataset('x', data=self.data)

    def test_ndim(self):
        """ Verify number of dimensions """
        self.assertEqual(self.dset.ndim, 0)

    def test_size(self):
        """ Verify size """
        self.assertEqual(self.dset.size, 1)

    def test_nbytes(self):
        """ Verify nbytes """
        self.assertEqual(self.dset.nbytes, self.data.dtype.itemsize)  # not sure if 'f' is always alias for 'f4'

    def test_shape(self):
        """ Verify shape """
        self.assertEqual(self.dset.shape, tuple())

    def test_ellipsis(self):
        """ Ellipsis -> scalar ndarray """
        out = self.dset[...]
        if not self.is_hsds:
            # TBD - fix for HSDS
            self.assertArrayEqual(out, self.data)

    def test_tuple(self):
        """ () -> bare item """
        out = self.dset[()]
        if not self.is_hsds:
            # TBD - fix for HSDS
            self.assertArrayEqual(out, self.data.item())

    def test_slice(self):
        """ slice -> ValueError """
        with self.assertRaises(ValueError):
            self.dset[0:4]

    def test_index(self):
        """ index -> ValueError """
        with self.assertRaises(ValueError):
            self.dset[0]

    # FIXME: NumPy has IndexError instead
    def test_indexlist(self):
        """ index list -> ValueError """
        with self.assertRaises(ValueError):
            self.dset[[1,2,5]]

    # FIXME: NumPy permits this
    def test_mask(self):
        """ mask -> ValueError """
        mask = np.array(True, dtype='bool')
        with self.assertRaises(ValueError):
            self.dset[mask]

    def test_fieldnames(self):
        """ field name -> ValueError (no fields) """
        with self.assertRaises(ValueError):
            self.dset['field']


class TestScalarCompound(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        filename = self.getFileName("dataset_testscalarcompound")
        print("filename:", filename)
        self.f = h5py.File(filename, 'w')
        if isinstance(self.f.id.id, str) and self.f.id.id.startswith("g-"):
            # array types not working with HSDS
            self.is_hsds = True
        else:
            self.is_hsds = False
        self.data = np.array((42.5, -118, "Hello"), dtype=[('a', 'f'), ('b', 'i'), ('c', '|S10')])
        #self.dset = self.f.create_dataset('x', data=self.data)
        self.dset = self.f.create_dataset('x', (), dtype=[('a', 'f'), ('b', 'i'), ('c', '|S10')])
        self.dset[...] =  (42.5, -118, "Hello")

    def test_ndim(self):
        """ Verify number of dimensions """
        self.assertEqual(self.dset.ndim, 0)

    def test_shape(self):
        """ Verify shape """
        self.assertEqual(self.dset.shape, tuple())

    def test_size(self):
        """ Verify size """
        self.assertEqual(self.dset.size, 1)

    def test_nbytes(self):
        """ Verify nbytes """
        self.assertEqual(self.dset.nbytes, self.data.dtype.itemsize)

    def test_ellipsis(self):
        """ Ellipsis -> scalar ndarray """
        out = self.dset[...]
        # assertArrayEqual doesn't work with compounds; do manually
        self.assertIsInstance(out, np.ndarray)
        if not self.is_hsds:
            # TBD: Fix for hsds
            self.assertEqual(out.shape, self.data.shape)
            self.assertEqual(out.dtype, self.data.dtype)

    def test_tuple(self):
        """ () -> np.void instance """
        out = self.dset[()]
        if not self.is_hsds:
            # TBD: Fix for HSDS
            self.assertIsInstance(out, np.void)
            self.assertEqual(out.dtype, self.data.dtype)

    def test_slice(self):
        """ slice -> ValueError """
        with self.assertRaises(ValueError):
            self.dset[0:4]

    def test_index(self):
        """ index -> ValueError """
        with self.assertRaises(ValueError):
            self.dset[0]

    # FIXME: NumPy has IndexError instead
    def test_indexlist(self):
        """ index list -> ValueError """
        with self.assertRaises(ValueError):
            self.dset[[1,2,5]]

    # FIXME: NumPy permits this
    def test_mask(self):
        """ mask -> ValueError  """
        mask = np.array(True, dtype='bool')
        with self.assertRaises(ValueError):
            self.dset[mask]

    # failed with earlier h5py versions
    def test_fieldnames(self):
        """ field name -> bare value """

        #TBD: fix when field access is supported in h5serv/hsds
        if config.get("use_h5py"):
            out = self.dset['a']
            self.assertIsInstance(out, np.float32)
            self.assertEqual(out, self.dset['a'])



class TestScalarArray(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        filename = self.getFileName("dataset_testscalararray")
        print("filename:", filename)
        self.f = h5py.File(filename, 'w')
        if isinstance(self.f.id.id, str) and self.f.id.id.startswith("g-"):
            # array types not working with HSDS
            self.is_hsds = True
        else:
            self.is_hsds = False
        self.dt = np.dtype('(3,2)f')
        self.data = np.array([(3.2, -119), (42, 99.8), (3.14, 0)], dtype='f')
        self.dset = self.f.create_dataset('x', (), dtype=self.dt)
        try:
            self.dset[...] = self.data
        except (IOError, OSError) as oe:
            #TBD this is failing on HSDS
            if not self.is_hsds:
                raise oe

    # FIXME: HSDS failure
    def test_ellipsis(self):
        """ Ellipsis -> ndarray promoted to underlying shape """
        out = self.dset[...]
        if not self.is_hsds:
            self.assertArrayEqual(out, self.data)

    # FIXME: HSDS failure
    def test_tuple(self):
        """ () -> same as ellipsis """
        out = self.dset[...]
        if not self.is_hsds:
            self.assertArrayEqual(out, self.data)

    def test_slice(self):
        """ slice -> ValueError """
        with self.assertRaises(ValueError):
            self.dset[0:4]

    def test_index(self):
        """ index -> ValueError """
        with self.assertRaises(ValueError):
            self.dset[0]

    def test_indexlist(self):
        """ index list -> ValueError """
        with self.assertRaises(ValueError):
            self.dset[[]]

    def test_mask(self):
        """ mask -> ValueError """
        mask = np.array(True, dtype='bool')
        with self.assertRaises(ValueError):
            self.dset[mask]

    def test_fieldnames(self):
        """ field name -> ValueError (no fields) """
        with self.assertRaises(ValueError):
            self.dset['field']


class Test1DZeroFloat(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        filename = self.getFileName("dataset_test1dzerofloat")
        print("filename:", filename)
        self.f = h5py.File(filename, 'w')
        self.data = np.ones((0,), dtype='f')
        self.dset = self.f.create_dataset('x', data=self.data)

    def test_ndim(self):
        """ Verify number of dimensions """
        self.assertEqual(self.dset.ndim, 1)

    def test_shape(self):
        """ Verify shape """
        self.assertEqual(self.dset.shape, (0,))

    def test_ellipsis(self):
        """ Ellipsis -> ndarray of matching shape """
        self.assertNumpyBehavior(self.dset, self.data, np.s_[...])

    def test_tuple(self):
        """ () -> same as ellipsis """
        self.assertNumpyBehavior(self.dset, self.data, np.s_[()])

    def test_slice(self):
        """ slice -> ndarray of shape (0,) """
        self.assertNumpyBehavior(self.dset, self.data, np.s_[0:4])

    def test_slice_stop_less_than_start(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[7:5])

    def test_index(self):
        """ index -> out of range """
        with self.assertRaises(IndexError):
            self.dset[0]

    def test_indexlist(self):
        """ index list """
        self.assertNumpyBehavior(
            self.dset,
            self.data,
            np.s_[[]]
        )

    def test_mask(self):
        """ mask -> ndarray of matching shape """
        mask = np.ones((0,), dtype='bool')
        self.assertNumpyBehavior(
            self.dset, self.data, np.s_[mask])

    def test_fieldnames(self):
        """ field name -> ValueError (no fields) """
        with self.assertRaises(ValueError):
            self.dset['field']


class Test1DFloat(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        filename = self.getFileName("dataset_test1dfloat")
        print("filename:", filename)
        self.f = h5py.File(filename, 'w')
        self.data = np.arange(13).astype('f')
        self.dset = self.f.create_dataset('x', data=self.data)
        self.dset[...] = self.data

    def test_ndim(self):
        """ Verify number of dimensions """
        self.assertEqual(self.dset.ndim, 1)

    def test_shape(self):
        """ Verify shape """
        self.assertEqual(self.dset.shape, (13,))

    def test_ellipsis(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[...])

    def test_tuple(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[()])

    def test_slice_simple(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[0:4])

    def test_slice_zerosize(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[4:4])

    def test_slice_strides(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[1:7:3])

    def test_slice_negindexes(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[-8:-2:3])

    def test_slice_stop_less_than_start(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[7:5])

    def test_slice_outofrange(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[100:400:3])

    def test_slice_backwards(self):
        """ we disallow negative steps """
        with self.assertRaises(ValueError):
            self.dset[::-1]

    def test_slice_zerostride(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[::0])

    def test_index_simple(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[3])

    def test_index_neg(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[-4])

    # FIXME: NumPy permits this... it adds a new axis in front
    def test_index_none(self):
        with self.assertRaises(TypeError):
            self.dset[None]

    def test_index_illegal(self):
        """ Illegal slicing argument """
        with self.assertRaises(TypeError):
            self.dset[{}]

    def test_index_outofrange(self):
        with self.assertRaises(IndexError):
            self.dset[100]

    def test_indexlist_simple(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[[1,2,5]])

    def test_indexlist_single_index_ellipsis(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[[0], ...])

    def test_indexlist_numpyarray_single_index_ellipsis(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[np.array([0]), ...])

    def test_indexlist_numpyarray_ellipsis(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[np.array([1, 2, 5]), ...])

    def test_indexlist_empty(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[[]])

    def test_indexlist_outofrange(self):
        with self.assertRaises(IndexError):
            self.dset[[100]]

    def test_indexlist_nonmonotonic(self):
        """ we require index list values to be strictly increasing """
        with self.assertRaises(TypeError):
            self.dset[[1,3,2]]

    def test_indexlist_repeated(self):
        """ we forbid repeated index values """
        with self.assertRaises(TypeError):
            self.dset[[1,1,2]]

    def test_mask_true(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[self.data > -100])

    def test_mask_false(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[self.data > 100])

    def test_mask_partial(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[self.data > 5])

    def test_mask_wrongsize(self):
        """ we require the boolean mask shape to match exactly """
        with self.assertRaises(TypeError):
            self.dset[np.ones((2,), dtype='bool')]

    def test_fieldnames(self):
        """ field name -> ValueError (no fields) """
        with self.assertRaises(ValueError):
            self.dset['field']


class Test2DZeroFloat(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        filename = self.getFileName("dataset_test2dzerofloat")
        print("filename:", filename)
        self.f = h5py.File(filename, 'w')
        self.data = np.ones((0,3), dtype='f')
        self.dset = self.f.create_dataset('x', data=self.data)

    def test_ndim(self):
        """ Verify number of dimensions """
        self.assertEqual(self.dset.ndim, 2)

    def test_shape(self):
        """ Verify shape """
        self.assertEqual(self.dset.shape, (0, 3))

    def test_indexlist(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[:,[0,1,2]])

class Test2DFloat(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        filename = self.getFileName("dataset_test2dfloat")
        print("filename:", filename)
        self.f = h5py.File(filename, 'w')
        self.data = np.ones((5,3), dtype='f')
        self.dset = self.f.create_dataset('x', data=self.data)
         
    def test_ndim(self):
        """ Verify number of dimensions """
        self.assertEqual(self.dset.ndim, 2)

    def test_size(self):
        """ Verify size """
        self.assertEqual(self.dset.size, 15)

    def test_nbytes(self):
        """ Verify nbytes """
        self.assertEqual(self.dset.nbytes, 15*self.data.dtype.itemsize) 

    def test_shape(self):
        """ Verify shape """
        self.assertEqual(self.dset.shape, (5, 3))

    def test_indexlist(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[:,[0,1,2]])

    @ut.expectedFailure
    def test_index_emptylist(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[:, []])
        # test with single empty list failing - but results seems compat
        # with h5py 3.2.1 at least
        self.assertNumpyBehavior(self.dset, self.data, np.s_[[]])

class Test3DFloat(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        filename = self.getFileName("dataset_test3dfloat")
        print("filename:", filename)
        self.f = h5py.File(filename, 'w')
        self.data = np.ones((4,6,8), dtype='f')
        self.dset = self.f.create_dataset('x', data=self.data, dtype='f')

    def test_index_simple(self):
        self.assertNumpyBehavior(self.dset, self.data, np.s_[1,2:4,3:6])

class TestVeryLargeArray(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        filename = self.getFileName("dataset_testverylargearray")
        print("filename:", filename)
        self.f = h5py.File(filename, 'w')
        self.dset = self.f.create_dataset('x', shape=(2**15, 2**16))

    @ut.skipIf(sys.maxsize < 2**31, 'Maximum integer size >= 2**31 required')
    def test_size(self):
        self.assertEqual(self.dset.size, 2**31)


if __name__ == '__main__':
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    ut.main()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

from __future__ import absolute_import

import sys
import os
import os.path as op
import tempfile
import time
import config


from six import unichr

import numpy as np

if sys.version_info >= (2, 7) or sys.version_info >= (3, 2):
    import unittest as ut
else:
    try:
        import unittest2 as ut
    except ImportError:
        raise ImportError(
            'unittest2 is required to run the test suite with python-%d.%d'
            % (sys.version_info[:2])
            )


# Check if non-ascii filenames are supported
# Evidently this is the most reliable way to check
# See also h5py issue #263 and ipython #466
# To test for this, run the testsuite with LC_ALL=C
try:
    testfile, fname = tempfile.mkstemp(unichr(0x03b7))
except UnicodeError:
    UNICODE_FILENAMES = False
else:
    UNICODE_FILENAMES = True
    os.close(testfile)
    os.unlink(fname)
    del fname
    del testfile


class TestCase(ut.TestCase):

    """
        Base class for unit tests.
    """

    @property
    def test_user1(self):
        # HS_USERNAME is the username h5pyd will look up if
        #   if not provided in the File constructor
        user1 = {}
        if  "HS_USERNAME" in os.environ:
            user1["name"] = os.environ["HS_USERNAME"]
        else:
            user1["name"] = "test_user1"
        if "HS_PASSWORD" in os.environ:
            user1["password"] = os.environ["HS_PASSWORD"]
        else:
            # only use "test_user1/test" for desktop testing
            user1["password"] = "test"
        return user1

    @property
    def test_user2(self):
        user2 = {}
        if  "TEST12_USERNAME" in os.environ:
            user2["name"] = os.environ["TEST2_USERNAME"]
        else:
            user2["name"] = "test_user2"
        if "TEST2_PASSWORD" in os.environ:
            user2["password"] = os.environ["TEST2_PASSWORD"]
        else:
            # only use "test_user1/test" for desktop testing
            user2["password"] = "test"
        return user2

    @classmethod
    def use_h5py():
        """ Use the standard H5PY package rather than h5pyd"""
        if "USE_H5PY" in os.environ and os.environ["USE_H5PY"]:
            return True
        else:
            return False

    @classmethod
    def setUpClass(cls):
        pass
        #cls.tempdir = tempfile.mkdtemp(prefix='h5py-test_')

    @classmethod
    def tearDownClass(cls):
        pass
        #shutil.rmtree(cls.tempdir)

    def setUp(self):
        self.test_dir = str(int(time.time()))
        #self.f = h5py.File(self.mktemp(), 'w')

    def tearDown(self):
        try:
            if self.f:
                self.f.close()
        except:
            pass

    if not hasattr(ut.TestCase, 'assertSameElements'):
        # shim until this is ported into unittest2
        def assertSameElements(self, a, b):
            for x in a:
                match = False
                for y in b:
                    if x == y:
                        match = True
                if not match:
                    raise AssertionError("Item '%s' appears in a but not b" % x)

            for x in b:
                match = False
                for y in a:
                    if x == y:
                        match = True
                if not match:
                    raise AssertionError("Item '%s' appears in b but not a" % x)

    def assertArrayEqual(self, dset, arr, message=None, precision=None):
        """ Make sure dset and arr have the same shape, dtype and contents, to
            within the given precision.

            Note that dset may be a NumPy array or an HDF5 dataset.
        """
        if precision is None:
            precision = 1e-5
        if message is None:
            message = ''
        else:
            message = ' (%s)' % message

        if np.isscalar(dset) or np.isscalar(arr):
            self.assertTrue(
                np.isscalar(dset) and np.isscalar(arr),
                'Scalar/array mismatch ("%r" vs "%r")%s' % (dset, arr, message)
            )
            self.assertTrue(
                dset - arr < precision,
                "Scalars differ by more than %.3f%s" % (precision, message)
            )
            return

        self.assertTrue(
            dset.shape == arr.shape,
            "Shape mismatch (%s vs %s)%s" % (dset.shape, arr.shape, message)
            )
        self.assertTrue(
            dset.dtype == arr.dtype,
            "Dtype mismatch (%s vs %s)%s" % (dset.dtype, arr.dtype, message)
            )

        if arr.dtype.names is not None:
            for n in arr.dtype.names:
                message = '[FIELD %s] %s' % (n, message)
                self.assertArrayEqual(dset[n], arr[n], message=message, precision=precision)
        elif arr.dtype.kind in ('i', 'f'):
            self.assertTrue(
                np.all(np.abs(dset[...] - arr[...]) < precision),
                "Arrays differ by more than %.3f%s" % (precision, message)
                )
        else:
            self.assertTrue(
                np.all(dset[...] == arr[...]),
                "Arrays are not equal (dtype %s) %s" % (arr.dtype.str, message)
                )

    def assertNumpyBehavior(self, dset, arr, s):
        """ Apply slicing arguments "s" to both dset and arr.

        Succeeds if the results of the slicing are identical, or the
        exception raised is of the same type for both.

        "arr" must be a Numpy array; "dset" may be a NumPy array or dataset.
        """
        exc = None
        try:
            arr_result = arr[s]
        except Exception as e:
            exc = type(e)

        if exc is None:
            self.assertArrayEqual(dset[s], arr_result)
        else:
            with self.assertRaises(exc):
                dset[s]

    def getFileName(self, basename):
        """
        Get filepath for a test case given a testname
        """

        if config.get("use_h5py"):
            if not op.isdir("out"):
                os.mkdir("out")
            filename = "out/" + basename + ".h5"
        else:
            if "H5PYD_TEST_FOLDER" in os.environ:
                domain = os.environ["H5PYD_TEST_FOLDER"]
            else:
                domain = "h5pyd_test.hdfgroup.org"
            if domain.find('/') > -1:
                # Use path-style domain naming
                filename = op.join(domain, basename)
                filename += ".h5"
            else:
                filename = basename + "." + domain
        return filename


    def getPathFromDomain(self, domain):
        """
        Convert DNS-style domain name to filepath
        E.g. "mytest.h5pyd_test.hdfgroup.org" to
             "/org/hdfgroup/h5pyd_test/mytest
        """
        if domain.find('/') > -1:
            # looks like the domain already is specified as a path
            return domain

        names = domain.split('.')
        names.reverse()
        path = '/'
        for name in names:
            if name:
                 path += name
                 path += '/'
        path = path[:-1]  # strip trailing slash
        return path





##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import numpy as np
import logging
import six
import config

if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

from common import ut, TestCase

# Note: this test currently works with the HSDS server but not h5serv

class TestVlenTypes(TestCase):


    def test_create_vlen_attr(self):
        filename = self.getFileName("create_vlen_attribute")
        print("filename:", filename)
        if config.get("use_h5py"):
            # TBD - skipping as this core dumps in travis for some reason
            return
        f = h5py.File(filename, 'w')
        if isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            # vlen ref types not working for h5serv, so abort here
            f.close()
            return

        g1 = f.create_group('g1')
        g1_1 = g1.create_group('g1_1')
        g1_1.attrs["name"] = 'g1_1'
        g1_2 = g1.create_group('g1_2')
        g1_2.attrs["name"] = 'g1_2'
        g1_3 = g1.create_group('g1_3')
        g1_3.attrs["name"] = 'g1_3'

        # create an attribute that is a VLEN int32
        dtvlen = h5py.special_dtype(vlen=np.dtype('int32'))
        e0 = np.array([0,1,2])
        e1 = np.array([0,1,2,3])
        data = np.array([e0, e1], dtype=object)

        g1.attrs.create("a1", data, shape=(2,), dtype=dtvlen)

        ret_val = g1.attrs["a1"]
        self.assertTrue(isinstance(ret_val, np.ndarray))
        self.assertEqual(len(ret_val), 2)
        self.assertTrue(isinstance(ret_val[0], np.ndarray))
        # py36  attribute[a1]: [array([0, 1, 2], dtype=int32) array([0, 1, 2, 3], dtype=int32)]
        # py27  [(0, 1, 2) (0, 1, 2, 3)]
        self.assertEqual(list(ret_val[0]), [0,1,2])
        self.assertEqual(ret_val[0].dtype, np.dtype('int32'))
        self.assertTrue(isinstance(ret_val[1], np.ndarray))
        self.assertEqual(ret_val[1].dtype, np.dtype('int32'))

        self.assertEqual(list(ret_val[1]), [0,1,2,3])

        # create an attribute that is VLEN ObjRef
        dtref = h5py.special_dtype(ref=h5py.Reference)
        dtvlen = h5py.special_dtype(vlen=dtref)
        e0 = np.array((g1_1.ref,), dtype=dtref)
        e1 = np.array((g1_1.ref,g1_2.ref), dtype=dtref)
        e2 = np.array((g1_1.ref,g1_2.ref,g1_3.ref), dtype=dtref)
        data = [e0,e1,e2]

        g1.attrs.create("b1", data, shape=(3,),dtype=dtvlen)

        vlen_val = g1.attrs["b1"]  # read back attribute
        self.assertTrue(isinstance(vlen_val, np.ndarray))
        self.assertEqual(len(vlen_val), 3)
        for i in range(3):
            e = vlen_val[i]
            self.assertTrue(isinstance(e, np.ndarray))
            ref_type = h5py.check_dtype(ref=e.dtype)
            self.assertEqual(ref_type, h5py.Reference)
            # TBD - h5pyd is returning shape of () rather than (1,) for singletons
            if i>0:
                self.assertEqual(e.shape, ((i+1),))
                # first element is always a ref to g1
                refd_group = f[e[0]]
                self.assertEqual(refd_group.attrs['name'], 'g1_1')

        # create an attribute with compound type of vlen objref and int32
        dtcompound = np.dtype([('refs', dtvlen), ('number', 'int32')])
        # create np array with data for the attribute
        # note: two step process is needed, see: https://github.com/h5py/h5py/issues/573
        data = np.zeros((2,), dtype=dtcompound)
        data[0] = (e1, 1)
        data[1] = (e2, 2)

        g1.attrs.create("c1", data, shape=(2,), dtype=dtcompound)
        compound_val = g1.attrs["c1"]
        self.assertTrue(isinstance(compound_val, np.ndarray))
        self.assertEqual(len(compound_val), 2)
        self.assertEqual(len(compound_val.dtype), 2)
        for i in range(2):
            item = compound_val[i]
            self.assertTrue(isinstance(item, np.void))
            self.assertEqual(len(item), 2)
            e = item[0]
            self.assertEqual(len(e), i+2)
            refd_group = f[e[0]]
            self.assertEqual(refd_group.attrs['name'], 'g1_1')
            self.assertEqual(item[1], i+1)

        # close file
        f.close()


    def test_create_vlen_dset(self):
        filename = self.getFileName("create_vlen_dset")
        print("filename:", filename)
        if config.get("use_h5py"):
            # TBD - skipping as this core dumps in travis for some reason
            return
        f = h5py.File(filename, 'w')

        if isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            # vlen ref types not working for h5serv, so abort here
            f.close()
            return

        g1 = f.create_group('g1')
        g1_1 = g1.create_group('g1_1')
        g1_1.attrs["name"] = 'g1_1'
        g1_2 = g1.create_group('g1_2')
        g1_2.attrs["name"] = 'g1_2'
        g1_3 = g1.create_group('g1_3')
        g1_3.attrs["name"] = 'g1_3'

        # create a dataset that is a VLEN int32
        dtvlen = h5py.special_dtype(vlen=np.dtype('uint16'))

        dset1 = f.create_dataset("dset1", shape=(2,), dtype=dtvlen)

        # create numpy object array
        e0 = np.array([1,2,3],dtype='uint16')
        e1 = np.array([1,2,3,4],dtype='uint16')
        data = np.array([e0, e1], dtype=dtvlen)

        # write data
        dset1[...] = data

        # read back data
        e = dset1[0]
        ret_val = dset1[...]
        self.assertTrue(isinstance(ret_val, np.ndarray))
        self.assertEqual(len(ret_val), 2)
        self.assertTrue(isinstance(ret_val[0], np.ndarray))
        # py36  attribute[a1]: [array([0, 1, 2], dtype=int32) array([0, 1, 2, 3], dtype=int32)]
        # py27  [(0, 1, 2) (0, 1, 2, 3)]
        self.assertEqual(list(ret_val[0]), [1,2,3])
        self.assertEqual(ret_val[0].dtype, np.dtype('uint16'))
        self.assertTrue(isinstance(ret_val[1], np.ndarray))
        self.assertEqual(ret_val[1].dtype, np.dtype('uint16'))

        self.assertEqual(list(ret_val[1]), [1,2,3,4])

        # Read back just one element
        e0 = dset1[0]
        self.assertEqual(len(e0), 3)
        self.assertEqual(list(e0), [1,2,3])
        
        # try writing int arrays into dataset
        data = [42,]  
        dset1[0] = data
        ret_val = dset1[...]
        self.assertEqual(list(ret_val[0]), [42])


        # TBD: Test for VLEN objref and comount as with attribute test above

        # close file
        f.close()

    def test_create_vlen_compound_dset(self):
        filename = self.getFileName("create_vlen_compound_dset")
        print("filename:", filename)
        f = h5py.File(filename, "w")

        count = 10
         # create a dataset that is a VLEN int32
        dtvlen = h5py.special_dtype(vlen=np.dtype('int32'))
        dt = np.dtype([('x', np.int32), ('vals', dtvlen)])
        dset = f.create_dataset('compound_vlen', (count,), dtype=dt)

        elem = dset[0]
        for i in range(count):
            elem['x'] = i
            elem['vals'] = np.array(list(range(i+1)), dtype=np.int32)
            dset[i] = elem

        e = dset[5]
        self.assertEqual(len(e), 2)
        self.assertEqual(e[0], 5)
        e1 = list(e[1])
        self.assertEqual(e1, list(range(6)))
        
        f.close()

    def test_create_vlen_2d_dset(self):
        filename = self.getFileName("create_vlen_2d_dset")
        print("filename:", filename)
        if config.get("use_h5py"):
            # TBD - skipping as this core dumps in travis for some reason
            return
        f = h5py.File(filename, 'w')

        if isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            # vlen ref types not working for h5serv, so abort here
            f.close()
            return

        # create a dataset that is a VLEN int32
        dtvlen = h5py.special_dtype(vlen=np.dtype('int32'))

        nrows = 2
        ncols = 3
        dset1 = f.create_dataset("dset1", shape=(nrows,ncols), dtype=dtvlen)

        # create numpy object array
        data = np.zeros((nrows,ncols), dtype=dtvlen)
        for i in range(nrows):
            for j in range(ncols):
                alist = []
                for k in range((i+1)*(j+1)):
                    alist.append(k)
                data[i,j] = np.array(alist, dtype="int32")

        # write data
        dset1[...] = data

        # read back data
        ret_val = dset1[...]
        self.assertTrue(isinstance(ret_val, np.ndarray))
        self.assertEqual(ret_val.shape, (nrows, ncols))
        e12 = ret_val[1,2]
        self.assertTrue(isinstance(e12, np.ndarray))
        # py36  attribute[a1]: [array([0, 1, 2], dtype=int32) array([0, 1, 2, 3], dtype=int32)]
        # py27  [(0, 1, 2) (0, 1, 2, 3)]
        self.assertEqual(list(e12), [0,1,2,3,4,5])
        self.assertEqual(e12.dtype, np.dtype('int32'))

        # Read back just one element
        e12 = dset1[1,2]
        self.assertTrue(isinstance(e12, np.ndarray))
        self.assertEqual(e12.shape, (6,))
        # py36  attribute[a1]: [array([0, 1, 2], dtype=int32) array([0, 1, 2, 3], dtype=int32)]
        # py27  [(0, 1, 2) (0, 1, 2, 3)]
        self.assertEqual(list(e12), [0,1,2,3,4,5])
        self.assertEqual(e12.dtype, np.dtype('int32'))

        # close file
        f.close()


    def test_variable_len_str_attr(self):
        filename = self.getFileName("variable_len_str_dset")
        print("filename:", filename)
        if config.get("use_h5py"):
            # TBD - skipping as this core dumps in travis for some reason
            return
        f = h5py.File(filename, "w")
        if isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            # vlen ref types not working for h5serv, so abort here
            f.close()
            return

        words = (b"one", b"two", b"three", b"four", b"five", b"six", b"seven", b"eight", b"nine", b"ten")

        dims = (10,)
        dt = h5py.special_dtype(vlen=bytes)
        f.attrs.create('a1', words, shape=dims, dtype=dt)


        vals = f.attrs["a1"]  # read back

        self.assertTrue("vlen" in vals.dtype.metadata)

        for i in range(10):
            self.assertEqual(vals[i], words[i])

        f.close()


    def test_variable_len_str_dset(self):
        filename = self.getFileName("variable_len_str_dset")
        print("filename:", filename)
        if config.get("use_h5py"):
            # TBD - skipping as this core dumps in travis for some reason
            return
        f = h5py.File(filename, "w")
        if isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            # vlen  types not working for h5serv, so abort here
            f.close()
            return

        dims = (10,)
        dt = h5py.special_dtype(vlen=bytes)
        dset = f.create_dataset('variable_len_str_dset', dims, dtype=dt)

        self.assertEqual(dset.name, "/variable_len_str_dset")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 1)
        self.assertEqual(dset.shape[0], 10)
        self.assertEqual(str(dset.dtype), 'object')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 1)
        self.assertEqual(dset.maxshape[0], 10)
        if config.get('use_h5py'):
            self.assertEqual(dset.fillvalue, None)
        else:
            self.assertEqual(dset.fillvalue, 0)

        self.assertEqual(dset[0], 0)

        words = (b"one", b"two", b"three", b"four", b"five", b"six", b"seven", b"eight", b"nine", b"ten")
        dset[:] = words
        vals = dset[:]  # read back

        self.assertTrue("vlen" in vals.dtype.metadata)

        for i in range(10):
            self.assertEqual(vals[i], words[i])

        f.close()

    def test_variable_len_unicode_dset(self):
        filename = self.getFileName("variable_len_unicode_dset")
        print("filename:", filename)
        """
        if config.get("use_h5py"):
            # TBD - skipping as this core dumps in travis for some reason
            return
        """
        f = h5py.File(filename, "w")
        if isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            f.close()
            return  # variable len types not working with h5serv

        dims = (10,)
        if six.PY2:
            dt = h5py.special_dtype(vlen=unicode)
        else:
            dt = h5py.special_dtype(vlen=str)

        dset = f.create_dataset('variable_len_unicode_dset', dims, dtype=dt)

        self.assertEqual(dset.name, "/variable_len_unicode_dset")
        self.assertTrue(isinstance(dset.shape, tuple))
        self.assertEqual(len(dset.shape), 1)
        self.assertEqual(dset.shape[0], 10)
        self.assertEqual(str(dset.dtype), 'object')
        self.assertTrue(isinstance(dset.maxshape, tuple))
        self.assertEqual(len(dset.maxshape), 1)
        self.assertEqual(dset.maxshape[0], 10)
        if config.get('use_h5py'):
            self.assertEqual(dset.fillvalue, None)
        else:
            self.assertEqual(dset.fillvalue, 0)

        # TBD: h5serv and hsds returning different values for null strings
        if config.get('use_h5py'):
            self.assertEqual(dset[0], '')
        else:
            # TBD: HSDS is converteing null values to zeros
            self.assertEqual(dset[0], 0)


        words = (u"one: \u4e00", u"two: \u4e8c", u"three: \u4e09", u"four: \u56db", u"five: \u4e94", u"six: \u516d", u"seven: \u4e03", u"eight: \u516b", u"nine: \u4e5d", u"ten: \u5341")
        dset[:] = words
        vals = dset[:]  # read back

        self.assertTrue("vlen" in vals.dtype.metadata)

        for i in range(10):
            self.assertEqual(vals[i], words[i])

        f.close()

    def test_variable_len_unicode_attr(self):
        filename = self.getFileName("variable_len_unicode_attr")
        print("filename:", filename)
        """
        if config.get("use_h5py"):
            # TBD - skipping as this core dumps in travis for some reason
            return
        """
        f = h5py.File(filename, "w")
        if isinstance(f.id.id, str) and not f.id.id.startswith("g-"):
            f.close()
            return  # variable len types not working with h5serv

        dims = (10,)
        if six.PY2:
            dt = h5py.special_dtype(vlen=unicode)
        else:
            dt = h5py.special_dtype(vlen=str)

        words = (u"one: \u4e00", u"two: \u4e8c", u"three: \u4e09", u"four: \u56db", u"five: \u4e94", u"six: \u516d", u"seven: \u4e03", u"eight: \u516b", u"nine: \u4e5d", u"ten: \u5341")

        f.attrs.create('a1', words, shape=dims, dtype=dt)

        vals = f.attrs["a1"]  # read back
        #print("type:", type(vals))
        self.assertTrue("vlen" in vals.dtype.metadata)

        for i in range(10):
            self.assertEqual(vals[i], words[i])
            if six.PY2:
                self.assertEqual(type(vals[i]), unicode)
            else:
                self.assertEqual(type(vals[i]), str)


        f.close()


if __name__ == '__main__':
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    ut.main()



#!/usr/bin/env python
##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including       s   #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import os
import sys
import config
import h5pyd

#
# Main
#

print("load_file.py")


if "H5PYD_TEST_FOLDER" not in os.environ:
    print("set H5PYD_TEST_FOLDER environment not set")
    sys.exit(1)

test_folder = os.environ["H5PYD_TEST_FOLDER"]

data_dir = "data"

s3_http_path = "https://s3.amazonaws.com/hdfgroup/data/hdf5test/"

parent =  h5pyd.Folder(test_folder)
filenames = config.get_test_filenames()

if not os.path.exists(data_dir):
    # make data directory for downloaded HDF5 files
    os.mkdir(data_dir)

for filename in filenames:
    print(filename)
    domain_path = os.path.join(test_folder, filename)
    print(domain_path)
    if filename in parent:
        print("found")
        continue
    # check to see if the file has already been downloaded
    hdf5_path = os.path.join(data_dir, filename)
    if not os.path.isfile(hdf5_path):
        # wget from S3
        s3path = s3_http_path + filename
        print("downloading", s3path)
        rc = os.system("wget -q https://s3.amazonaws.com/hdfgroup/data/hdf5test/{} -P {}".format(filename, data_dir))
        if rc != 0:
            sys.exit("Failed to retreive test data file")
    # run hsload for each file
    print("running hsload for {}".format(hdf5_path))
    rc = os.system("python ../../h5pyd/_apps/hsload.py {} {}".format(hdf5_path, test_folder))
    if rc != 0:
        sys.exit("Failed to load {}".format(filename))
print("load_files done")
    


##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
import config
import logging
if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

from common import ut, TestCase
from datetime import datetime
import os.path
import six

class TestGroup(TestCase):

    def test_create(self):
        # create main test file
        filename = self.getFileName("create_group")
        print("filename:", filename)
        f = h5py.File(filename, 'w')
        is_hsds = False
        if isinstance(f.id.id, str) and f.id.id.startswith("g-"):
            is_hsds = True  # HSDS has different permission defaults
        self.assertTrue('/' in f)
        r = f['/']

        self.assertEqual(len(r), 0)
        self.assertTrue(isinstance(r, h5py.Group))
        self.assertTrue(r.name, '/')
        self.assertEqual(len(r.attrs.keys()), 0)
        self.assertFalse('g1' in r)

        r.create_group("g1")
        self.assertEqual(len(r), 1)
        self.assertTrue('g1' in r)

        g1 = r['g1']
        self.assertTrue(g1.id.id != r.id.id)
        self.assertEqual(g1.name, "/g1")

        r.create_group("g1/g1.1")
        g1_1 = r["g1/g1.1"]
        self.assertEqual(g1_1.name, "/g1/g1.1")
        self.assertEqual(len(r), 1)
        self.assertEqual(len(g1), 1)

        r.create_group('g2')
        self.assertEqual(len(r), 2)
        keys = []
        # iterate through keys
        for k in r:
            keys.append(k)

        self.assertEqual(len(keys), 2)
        self.assertTrue('g1' in keys)
        self.assertTrue('g2' in keys)

        self.assertTrue('g1' in r)
        self.assertTrue('/g1' in r)
        g1 = r.get('/g1')
        self.assertTrue(g1.id.id)
        self.assertEqual(g1.name, '/g1')

        g1_class = r.get('g1', getclass=True)
        self.assertEqual(g1_class, h5py.Group)

        # try creating a group that already exists
        try:
            r.create_group('g1')
            self.assertTrue(False)
        except ValueError:
            pass # expected

        r.create_group('g3')
        self.assertEqual(len(r), 3)
        del r['g3']
        self.assertEqual(len(r), 2)

        r.require_group('g4')
        self.assertEqual(len(r), 3)
        r.require_group('g2')
        self.assertEqual(len(r), 3)

        # create a hardlink
        tmp_grp = r.create_group("tmp")
        r['g1.1'] = tmp_grp

        # try to replace the link
        try:
            r['g1.1'] = g1_1
            self.assertTrue(False)  # shouldn't get here'
        except RuntimeError:
            pass # expected

        del r['tmp']
        self.assertEqual(len(r), 4)

        # create a softlink
        r['mysoftlink'] = h5py.SoftLink('/g1/g1.1')
        self.assertTrue("mysoftlink" in r)
        self.assertEqual(len(r), 5)
        self.assertEqual(len(g1), 1)
        self.assertEqual(len(g1_1), 0)

        slink = r['mysoftlink']
        self.assertEqual(slink.id, g1_1.id)

        # create a file that we'll link to
        link_target_filename = self.getFileName("link_target")
        g = h5py.File(link_target_filename, 'w')
        g.create_group("somepath")
        g.close()

        # create a external hardlink
        r['myexternallink'] = h5py.ExternalLink(link_target_filename, "somepath")
        # test getclass
        g1_class = r.get('g1', getclass=True)
        self.assertEqual(g1_class, h5py.Group)
        linkee_class = r.get('mysoftlink', getclass=True)
        self.assertEqual(linkee_class, h5py.Group)
        link_class = r.get('mysoftlink', getclass=True, getlink=True)
        self.assertEqual(link_class, h5py.SoftLink)
        softlink = r.get('mysoftlink', getlink=True)
        self.assertEqual(softlink.path, '/g1/g1.1')

        linkee_class = r.get('myexternallink', getclass=True)
        link_class = r.get('myexternallink', getclass=True, getlink=True)
        self.assertEqual(link_class, h5py.ExternalLink)
        external_link = r.get('myexternallink', getlink=True)
        self.assertEqual(external_link.path, 'somepath')
        external_link_filename = external_link.filename
        if config.get('use_h5py') or is_hsds:
            self.assertTrue(external_link_filename.find('link_target') > -1)
        else:
            self.assertTrue(external_link_filename.find('link_target') > -1)
            # h5serv should be a DNS style name
            self.assertEqual(external_link_filename.find('/'), -1)

        links = r.items()
        got_external_link = False
        for link in links:
            title = link[0]
            obj = link[1]
            if title == 'myexternallink':
                self.assertTrue(obj is not None)
                self.assertEqual(len(obj), 0)
                self.assertTrue(obj.file.filename != filename)
                got_external_link = True

        self.assertTrue(got_external_link)

        del r['mysoftlink']
        self.assertEqual(len(r), 5)

        del r['myexternallink']
        self.assertEqual(len(r), 4)

        # create group using nested path
        g2 = r['g2']
        r['g1/g1.3'] = g2

        self.assertEqual(len(r), 4)

        # try creating a link with a space in the name
        r["a space"] = g2
        self.assertEqual(len(r), 5)

        # re-create softlink
        r['mysoftlink'] = h5py.SoftLink('/g1/g1.1')

        # Check group's last modified time
        if h5py.__name__ == "h5pyd":
            self.assertTrue(isinstance(g1.modified, datetime))
            #self.assertEqual(g1.modified.tzname(), six.u('UTC'))

        f.close()

        # re-open file in read-only mode
        f = h5py.File(filename, 'r')
        self.assertEqual(len(f), 6)
        for name in ("g1", "g2", "g4", "g1.1", "a space", "mysoftlink"):
            self.assertTrue(name in f)
        self.assertTrue("/g1/g1.1" in f)
        g1_1 = f["/g1/g1.1"]
        
        if is_hsds:
            linkee_class = r.get('mysoftlink', getclass=True)
            print("linkee_class:", linkee_class)
            # TBD: investigate why h5py returned None here
            self.assertEqual(linkee_class, h5py.Group)
            link_class = r.get('mysoftlink', getclass=True, getlink=True)
            self.assertEqual(link_class, h5py.SoftLink)
            softlink = r.get('mysoftlink', getlink=True)
            self.assertEqual(softlink.path, '/g1/g1.1')
        linked_obj = f["mysoftlink"]
        self.assertEqual(linked_obj.id, g1_1.id)
        f.close()




    def test_nested_create(self):
        filename = self.getFileName("create_nested_group")
        print("filename:", filename)
        f = h5py.File(filename, 'w')
        is_hsds = False
        if isinstance(f.id.id, str) and f.id.id.startswith("g-"):
            is_hsds = True  # HSDS has different permission defaults
        self.assertTrue('/' in f)
        r = f['/']
        self.assertEqual(len(r), 0)

        # create multiple groups in a path
        r.create_group("/g1/g1.1")
        self.assertEqual(len(r), 1)
        self.assertTrue("g1" in r)
        g1 = r["g1"]
        self.assertTrue("g1.1" in g1)

        # same think, but with relative path
        g1.create_group("g1.2/g1.2.1/g1.2.1.1")
        self.assertEqual(len(g1), 2)
        self.assertTrue("g1.2" in g1)

        f.close()


    def test_external_links(self):
        # create a file for use a link target
        if config.get("use_h5py"):
            # for some reason this test is failing in Travis
            return
        linked_filename = self.getFileName("linked_file")
        abs_filepath = os.path.abspath(linked_filename)
        if config.get("use_h5py"):
            rel_filepath = os.path.relpath(linked_filename)
        else:
            rel_filepath = "linked_file.h5"
        f = h5py.File(linked_filename, 'w')
        is_hsds = False
        if isinstance(f.id.id, str) and f.id.id.startswith("g-"):
            is_hsds = True
        g1 = f.create_group("g1")
        dset = g1.create_dataset('ds', (5,7), dtype='f4')
        dset_id = dset.id.id
        f.close()

        filename = self.getFileName("external_links")
        print("filename:", filename)

        f = h5py.File(filename, 'w')
        f["missing_link"] = h5py.ExternalLink(abs_filepath, "somepath")
        f["abspath_link"] = h5py.ExternalLink(abs_filepath, "/g1/ds")
        f["relpath_link"] = h5py.ExternalLink(rel_filepath, "/g1/ds")
        try:
            linked_obj = f["missing_link"]
            self.assertTrue(False)
        except KeyError:
            pass # expected

        try:
            linked_obj = f["abspath_link"]
            self.assertTrue(linked_obj.name, "/g1/ds")
            self.assertEqual(linked_obj.shape, (5, 7))
            # The following no longer works for h5py 2.8
            # self.assertEqual(linked_obj.id.id, dset_id)
        except KeyError:
            if config.get("use_h5py") or is_hsds:
                # absolute paths aren't working yet for h5serv
                self.assertTrue(False)

        linked_obj = f["relpath_link"]
        self.assertTrue(linked_obj.name, "/g1/ds")
        self.assertEqual(linked_obj.shape, (5, 7))
        if not config.get("use_h5py"):
            self.assertEqual(linked_obj.id.id, dset_id)

        f.close()

    def test_link_removal(self):

        def get_count(grp):
            count = 0
            for item in grp:
                count += 1
            return count
        # create a file for use a link target
        if config.get("use_h5py"):
            # for some reason this test is failing in Travis
            return
        filename = self.getFileName("test_link_removal")
        print(filename)

        f = h5py.File(filename, 'w')
        g1 = f.create_group("g1")
        dset = g1.create_dataset('ds', (5,7), dtype='f4')
        self.assertEqual(len(g1), 1)
        self.assertEqual(get_count(g1), 1)

        g1_clone = f["g1"]
        self.assertEqual(len(g1_clone), 1)
        self.assertEqual(get_count(g1_clone), 1)

        del g1["ds"]
        self.assertEqual(len(g1), 0)
        self.assertEqual(get_count(g1), 0)

        self.assertEqual(len(g1_clone), 0)
        self.assertEqual(get_count(g1_clone), 0)


        f.close()




if __name__ == '__main__':
    loglevel = logging.ERROR
    logging.basicConfig(format='%(asctime)s %(message)s', level=loglevel)
    ut.main()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
import sys
import h5pyd
import numpy as np

cube_side = 256
USER_NAME = "test_user1"
USER_PASSWD = "test"
ENDPOINT = "http://192.168.99.100:5101"
if len(sys.argv) > 1:
    if sys.argv[1] in ("-h", "--help"):
        print("Usage: python readdatacube [side] [-z]")
        sys.exit(1)
    cube_side = int(sys.argv[1])

filename = "cube_" + str(cube_side) + "_" + str(cube_side) + "_" + str(cube_side)


if len(sys.argv) > 2:
    if sys.argv[2] == "-z":
        filename += "_gz"


filename += ".h5pyd_test.hdfgroup.org"


print("filename:", filename)

f = h5pyd.File(filename, "r", username=USER_NAME, password=USER_PASSWD, endpoint=ENDPOINT)

print("name:", f.name)
print("uuid:", f.id.uuid)

print("get dataset")

dset = f['dset']

print("name:", dset.name)
print("uuid:", dset.id.uuid)
print("shape:", dset.shape)
print("dset.type:", dset.dtype)
print("dset.maxshape:", dset.maxshape)

print("reading data...")

stats = []
for i in range(cube_side):
    arr = dset[i,:,:]
    stats.append( (i, np.min(arr), np.max(arr), np.mean(arr),
                np.median(arr), np.std(arr) ) )

print("done!")

print("slice     min  max       mean     medien    stddev")
print("--------------------------------------------------")
for item in stats:
    print(item)

f.close()


